#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Variable subcommands."""
from __future__ import annotations

import json
import os
from json import JSONDecodeError

from sqlalchemy import select

from airflow.cli.simple_table import AirflowConsole
from airflow.cli.utils import print_export_output
from airflow.models import Variable
from airflow.utils import cli as cli_utils
from airflow.utils.cli import suppress_logs_and_warning
from airflow.utils.providers_configuration_loader import providers_configuration_loaded
from airflow.utils.session import create_session, provide_session


@suppress_logs_and_warning
@providers_configuration_loaded
def variables_list(args):
    """Display all the variables."""
    with create_session() as session:
        variables = session.scalars(select(Variable)).all()
    AirflowConsole().print_as(data=variables, output=args.output, mapper=lambda x: {"key": x.key})


@suppress_logs_and_warning
@providers_configuration_loaded
def variables_get(args):
    """Display variable by a given name."""
    try:
        if args.default is None:
            var = Variable.get(args.key, deserialize_json=args.json)
            print(var)
        else:
            var = Variable.get(args.key, deserialize_json=args.json, default_var=args.default)
            print(var)
    except (ValueError, KeyError) as e:
        raise SystemExit(str(e).strip("'\""))


@cli_utils.action_cli
@providers_configuration_loaded
def variables_set(args):
    """Create new variable with a given name, value and description."""
    Variable.set(args.key, args.value, args.description, serialize_json=args.json)
    print(f"Variable {args.key} created")


@cli_utils.action_cli
@providers_configuration_loaded
def variables_delete(args):
    """Delete variable by a given name."""
    Variable.delete(args.key)
    print(f"Variable {args.key} deleted")


@cli_utils.action_cli
@providers_configuration_loaded
@provide_session
def variables_import(args, session):
    """Import variables from a given file."""
    if not os.path.exists(args.file):
        raise SystemExit("Missing variables file.")
    with open(args.file) as varfile:
        try:
            var_json = json.load(varfile)
        except JSONDecodeError:
            raise SystemExit("Invalid variables file.")
    suc_count = fail_count = 0
    skipped = set()
    action_on_existing = args.action_on_existing_key
    existing_keys = set()
    if action_on_existing != "overwrite":
        existing_keys = set(session.scalars(select(Variable.key).where(Variable.key.in_(var_json))))
    if action_on_existing == "fail" and existing_keys:
        raise SystemExit(f"Failed. These keys: {sorted(existing_keys)} already exists.")
    for k, v in var_json.items():
        if action_on_existing == "skip" and k in existing_keys:
            skipped.add(k)
            continue
        try:
            Variable.set(k, v, serialize_json=not isinstance(v, str))
        except Exception as e:
            print(f"Variable import failed: {e!r}")
            fail_count += 1
        else:
            suc_count += 1
    print(f"{suc_count} of {len(var_json)} variables successfully updated.")
    if fail_count:
        print(f"{fail_count} variable(s) failed to be updated.")
    if skipped:
        print(
            f"The variables with these keys: {list(sorted(skipped))} "
            f"were skipped because they already exists"
        )


@providers_configuration_loaded
def variables_export(args):
    """Export all the variables to the file."""
    var_dict = {}
    with create_session() as session:
        qry = session.scalars(select(Variable))

        data = json.JSONDecoder()
        for var in qry:
            try:
                val = data.decode(var.val)
            except Exception:
                val = var.val
            var_dict[var.key] = val

    with args.file as varfile:
        json.dump(var_dict, varfile, sort_keys=True, indent=4)
        print_export_output("Variables", var_dict, varfile)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This package is deprecated. Please use `airflow.sensors` or `airflow.providers.*.sensors`."""
from __future__ import annotations

import warnings

from airflow.exceptions import RemovedInAirflow3Warning
from airflow.utils.deprecation_tools import add_deprecated_classes

warnings.warn(
    "This package is deprecated. Please use `airflow.sensors` or `airflow.providers.*.sensors`.",
    RemovedInAirflow3Warning,
    stacklevel=2,
)

__deprecated_classes = {
    "aws_athena_sensor": {
        "AthenaSensor": "airflow.providers.amazon.aws.sensors.athena.AthenaSensor",
    },
    "aws_glue_catalog_partition_sensor": {
        "AwsGlueCatalogPartitionSensor": (
            "airflow.providers.amazon.aws.sensors.glue_catalog_partition.GlueCatalogPartitionSensor"
        ),
    },
    "aws_redshift_cluster_sensor": {
        "AwsRedshiftClusterSensor": (
            "airflow.providers.amazon.aws.sensors.redshift_cluster.RedshiftClusterSensor"
        ),
    },
    "aws_sqs_sensor": {
        "SqsSensor": "airflow.providers.amazon.aws.sensors.sqs.SqsSensor",
        "SQSSensor": "airflow.providers.amazon.aws.sensors.sqs.SqsSensor",
    },
    "azure_cosmos_sensor": {
        "AzureCosmosDocumentSensor": (
            "airflow.providers.microsoft.azure.sensors.cosmos.AzureCosmosDocumentSensor"
        ),
    },
    "bash_sensor": {
        "STDOUT": "airflow.sensors.bash.STDOUT",
        "BashSensor": "airflow.sensors.bash.BashSensor",
        "Popen": "airflow.sensors.bash.Popen",
        "TemporaryDirectory": "airflow.sensors.bash.TemporaryDirectory",
        "gettempdir": "airflow.sensors.bash.gettempdir",
    },
    "bigquery_sensor": {
        "BigQueryTableExistenceSensor": (
            "airflow.providers.google.cloud.sensors.bigquery.BigQueryTableExistenceSensor"
        ),
        "BigQueryTableSensor": "airflow.providers.google.cloud.sensors.bigquery.BigQueryTableExistenceSensor",
    },
    "cassandra_record_sensor": {
        "CassandraRecordSensor": "airflow.providers.apache.cassandra.sensors.record.CassandraRecordSensor",
    },
    "cassandra_table_sensor": {
        "CassandraTableSensor": "airflow.providers.apache.cassandra.sensors.table.CassandraTableSensor",
    },
    "celery_queue_sensor": {
        "CeleryQueueSensor": "airflow.providers.celery.sensors.celery_queue.CeleryQueueSensor",
    },
    "datadog_sensor": {
        "DatadogSensor": "airflow.providers.datadog.sensors.datadog.DatadogSensor",
    },
    "file_sensor": {
        "FileSensor": "airflow.sensors.filesystem.FileSensor",
    },
    "ftp_sensor": {
        "FTPSensor": "airflow.providers.ftp.sensors.ftp.FTPSensor",
        "FTPSSensor": "airflow.providers.ftp.sensors.ftp.FTPSSensor",
    },
    "gcp_transfer_sensor": {
        "CloudDataTransferServiceJobStatusSensor":
            "airflow.providers.google.cloud.sensors.cloud_storage_transfer_service."
            "CloudDataTransferServiceJobStatusSensor",
        "GCPTransferServiceWaitForJobStatusSensor":
            "airflow.providers.google.cloud.sensors.cloud_storage_transfer_service."
            "CloudDataTransferServiceJobStatusSensor",
    },
    "gcs_sensor": {
        "GCSObjectExistenceSensor": "airflow.providers.google.cloud.sensors.gcs.GCSObjectExistenceSensor",
        "GCSObjectsWithPrefixExistenceSensor": (
            "airflow.providers.google.cloud.sensors.gcs.GCSObjectsWithPrefixExistenceSensor"
        ),
        "GCSObjectUpdateSensor": "airflow.providers.google.cloud.sensors.gcs.GCSObjectUpdateSensor",
        "GCSUploadSessionCompleteSensor": (
            "airflow.providers.google.cloud.sensors.gcs.GCSUploadSessionCompleteSensor"
        ),
        "GoogleCloudStorageObjectSensor": (
            "airflow.providers.google.cloud.sensors.gcs.GCSObjectExistenceSensor"
        ),
        "GoogleCloudStorageObjectUpdatedSensor": (
            "airflow.providers.google.cloud.sensors.gcs.GCSObjectUpdateSensor"
        ),
        "GoogleCloudStoragePrefixSensor": (
            "airflow.providers.google.cloud.sensors.gcs.GCSObjectsWithPrefixExistenceSensor"
        ),
        "GoogleCloudStorageUploadSessionCompleteSensor": (
            "airflow.providers.google.cloud.sensors.gcs.GCSUploadSessionCompleteSensor"
        ),
    },
    "imap_attachment_sensor": {
        "ImapAttachmentSensor": "airflow.providers.imap.sensors.imap_attachment.ImapAttachmentSensor",
    },
    "jira_sensor": {
        "JiraSensor": "airflow.providers.atlassian.jira.sensors.jira.JiraSensor",
        "JiraTicketSensor": "airflow.providers.atlassian.jira.sensors.jira.JiraTicketSensor",
    },
    "mongo_sensor": {
        "MongoSensor": "airflow.providers.mongo.sensors.mongo.MongoSensor",
    },
    "pubsub_sensor": {
        "PubSubPullSensor": "airflow.providers.google.cloud.sensors.pubsub.PubSubPullSensor",
    },
    "python_sensor": {
        "PythonSensor": "airflow.sensors.python.PythonSensor",
    },
    "redis_key_sensor": {
        "RedisKeySensor": "airflow.providers.redis.sensors.redis_key.RedisKeySensor",
    },
    "redis_pub_sub_sensor": {
        "RedisPubSubSensor": "airflow.providers.redis.sensors.redis_pub_sub.RedisPubSubSensor",
    },
    "sagemaker_training_sensor": {
        "SageMakerHook": "airflow.providers.amazon.aws.sensors.sagemaker.SageMakerHook",
        "SageMakerTrainingSensor": "airflow.providers.amazon.aws.sensors.sagemaker.SageMakerTrainingSensor",
    },
    "sftp_sensor": {
        "SFTPSensor": "airflow.providers.sftp.sensors.sftp.SFTPSensor",
    },
    "wasb_sensor": {
        "WasbBlobSensor": "airflow.providers.microsoft.azure.sensors.wasb.WasbBlobSensor",
        "WasbPrefixSensor": "airflow.providers.microsoft.azure.sensors.wasb.WasbPrefixSensor",
    },
    "weekday_sensor": {
        "DayOfWeekSensor": "airflow.sensors.weekday.DayOfWeekSensor",
    },
}

add_deprecated_classes(__deprecated_classes, __name__)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from airflow.api_connexion.exceptions import BadRequest
from airflow.auth.managers.fab.api_endpoints import role_and_permission_endpoint, user_endpoint
from airflow.www.extensions.init_auth_manager import get_auth_manager

if TYPE_CHECKING:
    from typing import Callable

    from airflow.api_connexion.types import APIResponse


def _require_fab(func: Callable) -> Callable:
    """
    Raise an HTTP error 400 if the auth manager is not FAB.

    Intended to decorate endpoints that have been migrated from Airflow API to FAB API.
    """

    def inner(*args, **kwargs):
        from airflow.auth.managers.fab.fab_auth_manager import FabAuthManager

        auth_mgr = get_auth_manager()
        if not isinstance(auth_mgr, FabAuthManager):
            raise BadRequest(
                detail="This endpoint is only available when using the default auth manager FabAuthManager."
            )
        else:
            warnings.warn(
                "This API endpoint is deprecated. "
                "Please use the API under /auth/fab/v1 instead for this operation.",
                DeprecationWarning,
            )
            return func(*args, **kwargs)

    return inner


### role


@_require_fab
def get_role(**kwargs) -> APIResponse:
    """Get role."""
    return role_and_permission_endpoint.get_role(**kwargs)


@_require_fab
def get_roles(**kwargs) -> APIResponse:
    """Get roles."""
    return role_and_permission_endpoint.get_roles(**kwargs)


@_require_fab
def delete_role(**kwargs) -> APIResponse:
    """Delete a role."""
    return role_and_permission_endpoint.delete_role(**kwargs)


@_require_fab
def patch_role(**kwargs) -> APIResponse:
    """Update a role."""
    kwargs.pop("body", None)
    return role_and_permission_endpoint.patch_role(**kwargs)


@_require_fab
def post_role(**kwargs) -> APIResponse:
    """Create a new role."""
    kwargs.pop("body", None)
    return role_and_permission_endpoint.post_role(**kwargs)


### permissions
@_require_fab
def get_permissions(**kwargs) -> APIResponse:
    """Get permissions."""
    return role_and_permission_endpoint.get_permissions(**kwargs)


### user
@_require_fab
def get_user(**kwargs) -> APIResponse:
    """Get a user."""
    return user_endpoint.get_user(**kwargs)


@_require_fab
def get_users(**kwargs) -> APIResponse:
    """Get users."""
    return user_endpoint.get_users(**kwargs)


@_require_fab
def post_user(**kwargs) -> APIResponse:
    """Create a new user."""
    kwargs.pop("body", None)
    return user_endpoint.post_user(**kwargs)


@_require_fab
def patch_user(**kwargs) -> APIResponse:
    """Update a user."""
    kwargs.pop("body", None)
    return user_endpoint.patch_user(**kwargs)


@_require_fab
def delete_user(**kwargs) -> APIResponse:
    """Delete a user."""
    return user_endpoint.delete_user(**kwargs)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Callable,
)

from fsspec.implementations.local import LocalFileSystem

from airflow.compat.functools import cache
from airflow.providers_manager import ProvidersManager
from airflow.stats import Stats
from airflow.utils.module_loading import import_string

if TYPE_CHECKING:
    from fsspec import AbstractFileSystem

    from airflow.io.typedef import Properties


log = logging.getLogger(__name__)


def _file(_: str | None, storage_options: Properties) -> LocalFileSystem:
    return LocalFileSystem(**storage_options)


# builtin supported filesystems
_BUILTIN_SCHEME_TO_FS: dict[str, Callable[[str | None, Properties], AbstractFileSystem]] = {
    "file": _file,
}


@cache
def _register_filesystems() -> dict[str, Callable[[str | None, Properties], AbstractFileSystem]]:
    scheme_to_fs = _BUILTIN_SCHEME_TO_FS.copy()
    with Stats.timer("airflow.io.load_filesystems") as timer:
        manager = ProvidersManager()
        for fs_module_name in manager.filesystem_module_names:
            fs_module = import_string(fs_module_name)
            for scheme in getattr(fs_module, "schemes", []):
                if scheme in scheme_to_fs:
                    log.warning("Overriding scheme %s for %s", scheme, fs_module_name)

                method = getattr(fs_module, "get_fs", None)
                if method is None:
                    raise ImportError(f"Filesystem {fs_module_name} does not have a get_fs method")
                scheme_to_fs[scheme] = method

    log.debug("loading filesystems from providers took %.3f seconds", timer.duration)
    return scheme_to_fs


def get_fs(
    scheme: str, conn_id: str | None = None, storage_options: Properties | None = None
) -> AbstractFileSystem:
    """
    Get a filesystem by scheme.

    :param scheme: the scheme to get the filesystem for
    :return: the filesystem method
    :param conn_id: the airflow connection id to use
    :param storage_options: the storage options to pass to the filesystem
    """
    filesystems = _register_filesystems()
    try:
        fs = filesystems[scheme]
    except KeyError:
        raise ValueError(f"No filesystem registered for scheme {scheme}") from None

    options = storage_options or {}
    return fs(conn_id, options)


def has_fs(scheme: str) -> bool:
    """
    Check if a filesystem is available for a scheme.

    :param scheme: the scheme to check
    :return: True if a filesystem is available for the scheme
    """
    return scheme in _register_filesystems()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""DagProcessor command."""
from __future__ import annotations

import logging
from datetime import timedelta
from typing import Any

from airflow.cli.commands.daemon_utils import run_command_with_daemon_option
from airflow.configuration import conf
from airflow.dag_processing.manager import DagFileProcessorManager, reload_configuration_for_dag_processing
from airflow.jobs.dag_processor_job_runner import DagProcessorJobRunner
from airflow.jobs.job import Job, run_job
from airflow.utils import cli as cli_utils
from airflow.utils.providers_configuration_loader import providers_configuration_loaded

log = logging.getLogger(__name__)


def _create_dag_processor_job_runner(args: Any) -> DagProcessorJobRunner:
    """Create DagFileProcessorProcess instance."""
    processor_timeout_seconds: int = conf.getint("core", "dag_file_processor_timeout")
    processor_timeout = timedelta(seconds=processor_timeout_seconds)
    return DagProcessorJobRunner(
        job=Job(),
        processor=DagFileProcessorManager(
            processor_timeout=processor_timeout,
            dag_directory=args.subdir,
            max_runs=args.num_runs,
            dag_ids=[],
            pickle_dags=args.do_pickle,
        ),
    )


@cli_utils.action_cli
@providers_configuration_loaded
def dag_processor(args):
    """Start Airflow Dag Processor Job."""
    if not conf.getboolean("scheduler", "standalone_dag_processor"):
        raise SystemExit("The option [scheduler/standalone_dag_processor] must be True.")

    sql_conn: str = conf.get("database", "sql_alchemy_conn").lower()
    if sql_conn.startswith("sqlite"):
        raise SystemExit("Standalone DagProcessor is not supported when using sqlite.")

    job_runner = _create_dag_processor_job_runner(args)

    reload_configuration_for_dag_processing()
    run_command_with_daemon_option(
        args=args,
        process_name="dag-processor",
        callback=lambda: run_job(job=job_runner.job, execute_callable=job_runner._execute),
        should_setup_logging=True,
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pathlib
from datetime import datetime, timedelta
from typing import Any, List, Optional

from dateutil import relativedelta
from pydantic import (
    BaseModel as BaseModelPydantic,
    ConfigDict,
    PlainSerializer,
    PlainValidator,
    ValidationInfo,
)
from typing_extensions import Annotated

from airflow import DAG, settings
from airflow.configuration import conf as airflow_conf
from airflow.utils.sqlalchemy import Interval


def serialize_interval(value: Interval) -> Interval:
    interval = Interval()
    return interval.process_bind_param(value, None)


def validate_interval(value: Interval | Any, _info: ValidationInfo) -> Any:
    if (
        isinstance(value, Interval)
        or isinstance(value, timedelta)
        or isinstance(value, relativedelta.relativedelta)
    ):
        return value
    interval = Interval()
    try:
        return interval.process_result_value(value, None)
    except ValueError as e:
        # Interval may be provided in string format (cron),
        # so it must be returned as valid value.
        if isinstance(value, str):
            return value
        raise e


PydanticInterval = Annotated[
    Interval,
    PlainValidator(validate_interval),
    PlainSerializer(serialize_interval, return_type=Interval),
]


def serialize_operator(x: DAG) -> dict:
    from airflow.serialization.serialized_objects import SerializedDAG

    return SerializedDAG.serialize_dag(x)


def validate_operator(x: DAG | dict[str, Any], _info: ValidationInfo) -> Any:
    from airflow.serialization.serialized_objects import SerializedDAG

    if isinstance(x, DAG):
        return x
    return SerializedDAG.deserialize_dag(x)


PydanticDag = Annotated[
    DAG,
    PlainValidator(validate_operator),
    PlainSerializer(serialize_operator, return_type=dict),
]


class DagOwnerAttributesPydantic(BaseModelPydantic):
    """Serializable representation of the DagOwnerAttributes ORM SqlAlchemyModel used by internal API."""

    owner: str
    link: str

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)


class DagTagPydantic(BaseModelPydantic):
    """Serializable representation of the DagTag ORM SqlAlchemyModel used by internal API."""

    name: str
    dag_id: str

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)


class DagModelPydantic(BaseModelPydantic):
    """Serializable representation of the DagModel ORM SqlAlchemyModel used by internal API."""

    dag_id: str
    root_dag_id: Optional[str]
    is_paused_at_creation: bool = airflow_conf.getboolean("core", "dags_are_paused_at_creation")
    is_paused: bool = is_paused_at_creation
    is_subdag: Optional[bool] = False
    is_active: Optional[bool] = False
    last_parsed_time: Optional[datetime]
    last_pickled: Optional[datetime]
    last_expired: Optional[datetime]
    scheduler_lock: Optional[bool]
    pickle_id: Optional[int]
    fileloc: str
    processor_subdir: Optional[str]
    owners: Optional[str]
    description: Optional[str]
    default_view: Optional[str]
    schedule_interval: Optional[PydanticInterval]
    timetable_description: Optional[str]
    tags: List[DagTagPydantic]  # noqa
    dag_owner_links: List[DagOwnerAttributesPydantic]  # noqa
    parent_dag: Optional[PydanticDag]

    max_active_tasks: int
    max_active_runs: Optional[int]

    has_task_concurrency_limits: bool
    has_import_errors: Optional[bool] = False

    _processor_dags_folder: Optional[str] = None

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)

    @property
    def relative_fileloc(self) -> pathlib.Path:
        """File location of the importable dag 'file' relative to the configured DAGs folder."""
        path = pathlib.Path(self.fileloc)
        try:
            rel_path = path.relative_to(self._processor_dags_folder or settings.DAGS_FOLDER)
            if rel_path == pathlib.Path("."):
                return path
            else:
                return rel_path
        except ValueError:
            # Not relative to DAGS_FOLDER.
            return path



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Add ``start_date`` and ``end_date`` in ``dag_run`` table

Revision ID: 4446e08588
Revises: 561833c1c74b
Create Date: 2015-12-10 11:26:18.439223

"""
from __future__ import annotations

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4446e08588"
down_revision = "561833c1c74b"
branch_labels = None
depends_on = None
airflow_version = "1.6.2"


def upgrade():
    op.add_column("dag_run", sa.Column("end_date", sa.DateTime(), nullable=True))
    op.add_column("dag_run", sa.Column("start_date", sa.DateTime(), nullable=True))


def downgrade():
    op.drop_column("dag_run", "start_date")
    op.drop_column("dag_run", "end_date")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Iterable, List, Optional

from pydantic import BaseModel as BaseModelPydantic, ConfigDict

from airflow.serialization.pydantic.dag import PydanticDag
from airflow.serialization.pydantic.dataset import DatasetEventPydantic
from airflow.utils.session import NEW_SESSION, provide_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from airflow.jobs.scheduler_job_runner import TI
    from airflow.serialization.pydantic.taskinstance import TaskInstancePydantic
    from airflow.utils.state import TaskInstanceState


class DagRunPydantic(BaseModelPydantic):
    """Serializable representation of the DagRun ORM SqlAlchemyModel used by internal API."""

    id: int
    dag_id: str
    queued_at: Optional[datetime]
    execution_date: datetime
    start_date: Optional[datetime]
    end_date: Optional[datetime]
    state: str
    run_id: str
    creating_job_id: Optional[int]
    external_trigger: bool
    run_type: str
    conf: dict
    data_interval_start: Optional[datetime]
    data_interval_end: Optional[datetime]
    last_scheduling_decision: Optional[datetime]
    dag_hash: Optional[str]
    updated_at: Optional[datetime]
    dag: Optional[PydanticDag]
    consumed_dataset_events: List[DatasetEventPydantic]  # noqa
    log_template_id: Optional[int]

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)

    @property
    def logical_date(self) -> datetime:
        return self.execution_date

    @provide_session
    def get_task_instances(
        self,
        state: Iterable[TaskInstanceState | None] | None = None,
        session: Session = NEW_SESSION,
    ) -> list[TI]:
        """
        Return the task instances for this dag run.

        TODO: make it works for AIP-44
        """
        raise NotImplementedError()

    @provide_session
    def get_task_instance(
        self,
        task_id: str,
        session: Session = NEW_SESSION,
        *,
        map_index: int = -1,
    ) -> TI | TaskInstancePydantic | None:
        """
        Return the task instance specified by task_id for this dag run.

        :param task_id: the task id
        :param session: Sqlalchemy ORM Session
        """
        from airflow.models.dagrun import DagRun

        return DagRun.fetch_task_instance(
            dag_id=self.dag_id,
            dag_run_id=self.run_id,
            task_id=task_id,
            session=session,
            map_index=map_index,
        )


DagRunPydantic.model_rebuild()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""DAG demonstrating various options for a trigger form generated by DAG params.

The DAG attribute `params` is used to define a default dictionary of parameters which are usually passed
to the DAG and which are used to render a trigger form.
"""
from __future__ import annotations

import datetime
import json
from pathlib import Path

from airflow.decorators import task
from airflow.models.dag import DAG
from airflow.models.param import Param, ParamsDict

with DAG(
    dag_id=Path(__file__).stem,
    description=__doc__.partition(".")[0],
    doc_md=__doc__,
    schedule=None,
    start_date=datetime.datetime(2022, 3, 4),
    catchup=False,
    tags=["example_ui"],
    params={
        # Let's start simple: Standard dict values are detected from type and offered as entry form fields.
        # Detected types are numbers, text, boolean, lists and dicts.
        # Note that such auto-detected parameters are treated as optional (not required to contain a value)
        "x": 3,
        "text": "Hello World!",
        "flag": False,
        "a_simple_list": ["one", "two", "three", "actually one value is made per line"],
        # But of course you might want to have it nicer! Let's add some description to parameters.
        # Note if you can add any Markdown formatting to the description, you need to use the description_md
        # attribute.
        "most_loved_number": Param(
            42,
            type="integer",
            title="Your favorite number",
            description_md="Everybody should have a **favorite** number. Not only _math teachers_. "
            "If you can not think of any at the moment please think of the 42 which is very famous because"
            "of the book [The Hitchhiker's Guide to the Galaxy]"
            "(https://en.wikipedia.org/wiki/Phrases_from_The_Hitchhiker%27s_Guide_to_the_Galaxy#"
            "The_Answer_to_the_Ultimate_Question_of_Life,_the_Universe,_and_Everything_is_42).",
        ),
        # If you want to have a selection list box then you can use the enum feature of JSON schema
        "pick_one": Param(
            "value 42",
            type="string",
            title="Select one Value",
            description="You can use JSON schema enum's to generate drop down selection boxes.",
            enum=[f"value {i}" for i in range(16, 64)],
        ),
        # You can also label the selected values via values_display attribute
        "pick_with_label": Param(
            3,
            type="number",
            title="Select one Number",
            description="With drop down selections you can also have nice display labels for the values.",
            enum=[*range(1, 10)],
            values_display={
                1: "One",
                2: "Two",
                3: "Three",
                4: "Four - is like you take three and get one for free!",
                5: "Five",
                6: "Six",
                7: "Seven",
                8: "Eight",
                9: "Nine",
            },
        ),
        # If you want to have a list box with proposals but not enforcing a fixed list
        # then you can use the examples feature of JSON schema
        "proposals": Param(
            "some value",
            type="string",
            title="Field with proposals",
            description="You can use JSON schema examples's to generate drop down selection boxes "
            "but allow also to enter custom values. Try typing an 'a' and see options.",
            examples=(
                "Alpha,Bravo,Charlie,Delta,Echo,Foxtrot,Golf,Hotel,India,Juliett,Kilo,Lima,Mike,November,Oscar,Papa,"
                "Quebec,Romeo,Sierra,Tango,Uniform,Victor,Whiskey,X-ray,Yankee,Zulu"
            ).split(","),
        ),
        # If you want to select multiple items from a fixed list JSON schema des not allow to use enum
        # In this case the type "array" is being used together with "examples" as pick list
        "multi_select": Param(
            ["two", "three"],
            "Select from the list of options.",
            type="array",
            title="Multi Select",
            examples=["one", "two", "three", "four", "five"],
        ),
        # A multiple options selection can also be combined with values_display
        "multi_select_with_label": Param(
            ["2", "3"],
            "Select from the list of options. See that options can have nicer text and still technical values"
            "are propagated as values during trigger to the DAG.",
            type="array",
            title="Multi Select with Labels",
            examples=["1", "2", "3", "4", "5"],
            values_display={
                "1": "One box of choccolate",
                "2": "Two bananas",
                "3": "Three apples",
                # Note: Value display mapping does not need to be complete.s
            },
        ),
        # An array of numbers
        "array_of_numbers": Param(
            [1, 2, 3],
            "Only integers are accepted in this array",
            type="array",
            title="Array of numbers",
            items={"type": "number"},
        ),
        # Boolean as proper parameter with description
        "bool": Param(
            True,
            type="boolean",
            title="Please confirm",
            description="A On/Off selection with a proper description.",
        ),
        # Dates and Times are also supported
        "date_time": Param(
            f"{datetime.date.today()}T{datetime.time(hour=12, minute=17, second=00)}+00:00",
            type="string",
            format="date-time",
            title="Date-Time Picker",
            description="Please select a date and time, use the button on the left for a pup-up calendar.",
        ),
        "date": Param(
            f"{datetime.date.today()}",
            type="string",
            format="date",
            title="Date Picker",
            description="Please select a date, use the button on the left for a pup-up calendar. "
            "See that here are no times!",
        ),
        "time": Param(
            f"{datetime.time(hour=12, minute=13, second=14)}",
            type=["string", "null"],
            format="time",
            title="Time Picker",
            description="Please select a time, use the button on the left for a pup-up tool.",
        ),
        # Fields can be required or not. If the defined fields are typed they are getting required by default
        # (else they would not pass JSON schema validation) - to make typed fields optional you must
        # permit the optional "null" type.
        # You can omit a default value if the DAG is triggered manually
        "required_field": Param(
            # In this example we have no default value
            # Form will enforce a value supplied by users to be able to trigger
            type="string",
            title="Required text field",
            description="This field is required. You can not submit without having text in here.",
        ),
        "optional_field": Param(
            "optional text, you can trigger also w/o text",
            type=["null", "string"],
            title="Optional text field",
            description_md="This field is optional. As field content is JSON schema validated you must "
            "allow the `null` type.",
        ),
        # You can arrange the entry fields in sections so that you can have a better overview for the user
        # Therefore you can add the "section" attribute.
        # The benefit of the Params class definition is that the full scope of JSON schema validation
        # can be leveraged for form fields and they will be validated before DAG submission.
        "checked_text": Param(
            "length-checked-field",
            type="string",
            title="Text field with length check",
            description_md="""This field is required. And you need to provide something between 10 and 30
            characters. See the JSON
            [schema description (string)](https://json-schema.org/understanding-json-schema/reference/string.html)
            for more details""",
            minLength=10,
            maxLength=20,
            section="JSON Schema validation options",
        ),
        "checked_number": Param(
            100,
            type="number",
            title="Number field with value check",
            description_md="""This field is required. You need to provide any number between 64 and 128.
            See the JSON
            [schema description (numbers)](https://json-schema.org/understanding-json-schema/reference/numeric.html)
            for more details""",
            minimum=64,
            maximum=128,
            section="JSON Schema validation options",
        ),
        # Some further cool stuff as advanced options are also possible
        # You can have the user entering a dict object as a JSON with validation
        "object": Param(
            {"key": "value"},
            type=["object", "null"],
            title="JSON entry field",
            section="Special advanced stuff with form fields",
        ),
        "array_of_objects": Param(
            [{"name": "account_name", "country": "country_name"}],
            description_md="Array with complex objects and validation rules. "
            "See [JSON Schema validation options in specs]"
            "(https://json-schema.org/understanding-json-schema/reference/array.html#items).",
            type="array",
            title="JSON array field",
            items={
                "type": "object",
                "properties": {"name": {"type": "string"}, "country_name": {"type": "string"}},
                "required": ["name"],
            },
            section="Special advanced stuff with form fields",
        ),
        # If you want to have static parameters which are always passed and not editable by the user
        # then you can use the JSON schema option of passing constant values. These parameters
        # will not be displayed but passed to the DAG
        "hidden_secret_field": Param("constant value", const="constant value"),
    },
) as dag:

    @task
    def show_params(**kwargs) -> None:
        params: ParamsDict = kwargs["params"]
        print(f"This DAG was triggered with the following parameters:\n\n{json.dumps(params, indent=4)}\n")

    show_params()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# mypy: disable-error-code=var-annotated
from __future__ import annotations

import logging

# This product contains a modified portion of 'Flask App Builder' developed by Daniel Vaz Gaspar.
# (https://github.com/dpgaspar/Flask-AppBuilder).
# Copyright 2013, Daniel Vaz Gaspar
log = logging.getLogger(__name__)

__lazy_imports = {
    "AUTH_DB": "flask_appbuilder.const",
    "AUTH_LDAP": "flask_appbuilder.const",
    "LOGMSG_WAR_SEC_LOGIN_FAILED": "flask_appbuilder.const",
}


def __getattr__(name: str):
    # PEP-562: Lazy loaded attributes on python modules
    path = __lazy_imports.get(name)
    if not path:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

    from airflow.utils.module_loading import import_string

    val = import_string(f"{path}.{name}")
    # Store for next time
    globals()[name] = val
    return val



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This package is deprecated. Please use `airflow.hooks` or `airflow.providers.*.hooks`."""
from __future__ import annotations

import warnings

from airflow.exceptions import RemovedInAirflow3Warning
from airflow.utils.deprecation_tools import add_deprecated_classes

warnings.warn(
    "This package is deprecated. Please use `airflow.hooks` or `airflow.providers.*.hooks`.",
    RemovedInAirflow3Warning,
    stacklevel=2,
)

__deprecated_classes = {
    "aws_athena_hook": {
        "AWSAthenaHook": "airflow.providers.amazon.aws.hooks.athena.AthenaHook",
    },
    "aws_datasync_hook": {
        "AWSDataSyncHook": "airflow.providers.amazon.aws.hooks.datasync.DataSyncHook",
    },
    "aws_dynamodb_hook": {
        "AwsDynamoDBHook": "airflow.providers.amazon.aws.hooks.dynamodb.DynamoDBHook",
    },
    "aws_firehose_hook": {
        "FirehoseHook": "airflow.providers.amazon.aws.hooks.kinesis.FirehoseHook",
    },
    "aws_glue_catalog_hook": {
        "AwsGlueCatalogHook": "airflow.providers.amazon.aws.hooks.glue_catalog.GlueCatalogHook",
    },
    "aws_hook": {
        "AwsBaseHook": "airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook",
        "_parse_s3_config": "airflow.providers.amazon.aws.hooks.base_aws._parse_s3_config",
        "boto3": "airflow.providers.amazon.aws.hooks.base_aws.boto3",
        "AwsHook": "airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook",
    },
    "aws_lambda_hook": {
        "AwsLambdaHook": "airflow.providers.amazon.aws.hooks.lambda_function.LambdaHook",
    },
    "aws_logs_hook": {
        "AwsLogsHook": "airflow.providers.amazon.aws.hooks.logs.AwsLogsHook",
    },
    "aws_sns_hook": {
        "AwsSnsHook": "airflow.providers.amazon.aws.hooks.sns.SnsHook",
    },
    "aws_sqs_hook": {
        "SqsHook": "airflow.providers.amazon.aws.hooks.sqs.SqsHook",
        "SQSHook": "airflow.providers.amazon.aws.hooks.sqs.SqsHook",
    },
    "azure_container_instance_hook": {
        "AzureContainerInstanceHook": (
            "airflow.providers.microsoft.azure.hooks.container_instance.AzureContainerInstanceHook"
        ),
    },
    "azure_container_registry_hook": {
        "AzureContainerRegistryHook": (
            "airflow.providers.microsoft.azure.hooks.container_registry.AzureContainerRegistryHook"
        ),
    },
    "azure_container_volume_hook": {
        "AzureContainerVolumeHook": (
            "airflow.providers.microsoft.azure.hooks.container_volume.AzureContainerVolumeHook"
        ),
    },
    "azure_cosmos_hook": {
        "AzureCosmosDBHook": "airflow.providers.microsoft.azure.hooks.cosmos.AzureCosmosDBHook",
    },
    "azure_data_lake_hook": {
        "AzureDataLakeHook": "airflow.providers.microsoft.azure.hooks.data_lake.AzureDataLakeHook",
    },
    "azure_fileshare_hook": {
        "AzureFileShareHook": "airflow.providers.microsoft.azure.hooks.fileshare.AzureFileShareHook",
    },
    "bigquery_hook": {
        "BigQueryBaseCursor": "airflow.providers.google.cloud.hooks.bigquery.BigQueryBaseCursor",
        "BigQueryConnection": "airflow.providers.google.cloud.hooks.bigquery.BigQueryConnection",
        "BigQueryCursor": "airflow.providers.google.cloud.hooks.bigquery.BigQueryCursor",
        "BigQueryHook": "airflow.providers.google.cloud.hooks.bigquery.BigQueryHook",
        "GbqConnector": "airflow.providers.google.cloud.hooks.bigquery.GbqConnector",
    },
    "cassandra_hook": {
        "CassandraHook": "airflow.providers.apache.cassandra.hooks.cassandra.CassandraHook",
    },
    "cloudant_hook": {
        "CloudantHook": "airflow.providers.cloudant.hooks.cloudant.CloudantHook",
    },
    "databricks_hook": {
        "CANCEL_RUN_ENDPOINT": "airflow.providers.databricks.hooks.databricks.CANCEL_RUN_ENDPOINT",
        "GET_RUN_ENDPOINT": "airflow.providers.databricks.hooks.databricks.GET_RUN_ENDPOINT",
        "RESTART_CLUSTER_ENDPOINT": "airflow.providers.databricks.hooks.databricks.RESTART_CLUSTER_ENDPOINT",
        "RUN_LIFE_CYCLE_STATES": "airflow.providers.databricks.hooks.databricks.RUN_LIFE_CYCLE_STATES",
        "RUN_NOW_ENDPOINT": "airflow.providers.databricks.hooks.databricks.RUN_NOW_ENDPOINT",
        "START_CLUSTER_ENDPOINT": "airflow.providers.databricks.hooks.databricks.START_CLUSTER_ENDPOINT",
        "SUBMIT_RUN_ENDPOINT": "airflow.providers.databricks.hooks.databricks.SUBMIT_RUN_ENDPOINT",
        "TERMINATE_CLUSTER_ENDPOINT": (
            "airflow.providers.databricks.hooks.databricks.TERMINATE_CLUSTER_ENDPOINT"
        ),
        "DatabricksHook": "airflow.providers.databricks.hooks.databricks.DatabricksHook",
        "RunState": "airflow.providers.databricks.hooks.databricks.RunState",
    },
    "datadog_hook": {
        "DatadogHook": "airflow.providers.datadog.hooks.datadog.DatadogHook",
    },
    "datastore_hook": {
        "DatastoreHook": "airflow.providers.google.cloud.hooks.datastore.DatastoreHook",
    },
    "dingding_hook": {
        "DingdingHook": "airflow.providers.dingding.hooks.dingding.DingdingHook",
        "requests": "airflow.providers.dingding.hooks.dingding.requests",
    },
    "discord_webhook_hook": {
        "DiscordWebhookHook": "airflow.providers.discord.hooks.discord_webhook.DiscordWebhookHook",
    },
    "emr_hook": {
        "EmrHook": "airflow.providers.amazon.aws.hooks.emr.EmrHook",
    },
    "fs_hook": {
        "FSHook": "airflow.hooks.filesystem.FSHook",
    },
    "ftp_hook": {
        "FTPHook": "airflow.providers.ftp.hooks.ftp.FTPHook",
        "FTPSHook": "airflow.providers.ftp.hooks.ftp.FTPSHook",
    },
    "gcp_api_base_hook": {
        "GoogleBaseHook": "airflow.providers.google.common.hooks.base_google.GoogleBaseHook",
        "GoogleCloudBaseHook": "airflow.providers.google.common.hooks.base_google.GoogleBaseHook",
    },
    "gcp_bigtable_hook": {
        "BigtableHook": "airflow.providers.google.cloud.hooks.bigtable.BigtableHook",
    },
    "gcp_cloud_build_hook": {
        "CloudBuildHook": "airflow.providers.google.cloud.hooks.cloud_build.CloudBuildHook",
    },
    "gcp_compute_hook": {
        "ComputeEngineHook": "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook",
        "GceHook": "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook",
    },
    "gcp_container_hook": {
        "GKEHook": "airflow.providers.google.cloud.hooks.kubernetes_engine.GKEHook",
        "GKEClusterHook": "airflow.providers.google.cloud.hooks.kubernetes_engine.GKEHook",
    },
    "gcp_dataflow_hook": {
        "DataflowHook": "airflow.providers.google.cloud.hooks.dataflow.DataflowHook",
        "DataFlowHook": "airflow.providers.google.cloud.hooks.dataflow.DataflowHook",
    },
    "gcp_dataproc_hook": {
        "DataprocHook": "airflow.providers.google.cloud.hooks.dataproc.DataprocHook",
        "DataProcHook": "airflow.providers.google.cloud.hooks.dataproc.DataprocHook",
    },
    "gcp_dlp_hook": {
        "CloudDLPHook": "airflow.providers.google.cloud.hooks.dlp.CloudDLPHook",
        "DlpJob": "airflow.providers.google.cloud.hooks.dlp.DlpJob",
    },
    "gcp_function_hook": {
        "CloudFunctionsHook": "airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook",
        "GcfHook": "airflow.providers.google.cloud.hooks.functions.CloudFunctionsHook",
    },
    "gcp_kms_hook": {
        "CloudKMSHook": "airflow.providers.google.cloud.hooks.kms.CloudKMSHook",
        "GoogleCloudKMSHook": "airflow.providers.google.cloud.hooks.kms.CloudKMSHook",
    },
    "gcp_mlengine_hook": {
        "MLEngineHook": "airflow.providers.google.cloud.hooks.mlengine.MLEngineHook",
    },
    "gcp_natural_language_hook": {
        "CloudNaturalLanguageHook": (
            "airflow.providers.google.cloud.hooks.natural_language.CloudNaturalLanguageHook"
        ),
    },
    "gcp_pubsub_hook": {
        "PubSubException": "airflow.providers.google.cloud.hooks.pubsub.PubSubException",
        "PubSubHook": "airflow.providers.google.cloud.hooks.pubsub.PubSubHook",
    },
    "gcp_spanner_hook": {
        "SpannerHook": "airflow.providers.google.cloud.hooks.spanner.SpannerHook",
        "CloudSpannerHook": "airflow.providers.google.cloud.hooks.spanner.SpannerHook",
    },
    "gcp_speech_to_text_hook": {
        "CloudSpeechToTextHook": "airflow.providers.google.cloud.hooks.speech_to_text.CloudSpeechToTextHook",
        "GCPSpeechToTextHook": "airflow.providers.google.cloud.hooks.speech_to_text.CloudSpeechToTextHook",
    },
    "gcp_sql_hook": {
        "CloudSQLDatabaseHook": "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook",
        "CloudSQLHook": "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook",
        "CloudSqlDatabaseHook": "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLDatabaseHook",
        "CloudSqlHook": "airflow.providers.google.cloud.hooks.cloud_sql.CloudSQLHook",
    },
    "gcp_tasks_hook": {
        "CloudTasksHook": "airflow.providers.google.cloud.hooks.tasks.CloudTasksHook",
    },
    "gcp_text_to_speech_hook": {
        "CloudTextToSpeechHook": "airflow.providers.google.cloud.hooks.text_to_speech.CloudTextToSpeechHook",
        "GCPTextToSpeechHook": "airflow.providers.google.cloud.hooks.text_to_speech.CloudTextToSpeechHook",
    },
    "gcp_transfer_hook": {
        "CloudDataTransferServiceHook": (
            "airflow.providers.google.cloud.hooks.cloud_storage_transfer_service.CloudDataTransferServiceHook"
        ),
        "GCPTransferServiceHook": (
            "airflow.providers.google.cloud.hooks.cloud_storage_transfer_service.CloudDataTransferServiceHook"
        ),
    },
    "gcp_translate_hook": {
        "CloudTranslateHook": "airflow.providers.google.cloud.hooks.translate.CloudTranslateHook",
    },
    "gcp_video_intelligence_hook": {
        "CloudVideoIntelligenceHook": (
            "airflow.providers.google.cloud.hooks.video_intelligence.CloudVideoIntelligenceHook"
        ),
    },
    "gcp_vision_hook": {
        "CloudVisionHook": "airflow.providers.google.cloud.hooks.vision.CloudVisionHook",
    },
    "gcs_hook": {
        "GCSHook": "airflow.providers.google.cloud.hooks.gcs.GCSHook",
        "GoogleCloudStorageHook": "airflow.providers.google.cloud.hooks.gcs.GCSHook",
    },
    "gdrive_hook": {
        "GoogleDriveHook": "airflow.providers.google.suite.hooks.drive.GoogleDriveHook",
    },
    "grpc_hook": {
        "GrpcHook": "airflow.providers.grpc.hooks.grpc.GrpcHook",
    },
    "imap_hook": {
        "ImapHook": "airflow.providers.imap.hooks.imap.ImapHook",
        "Mail": "airflow.providers.imap.hooks.imap.Mail",
        "MailPart": "airflow.providers.imap.hooks.imap.MailPart",
    },
    "jenkins_hook": {
        "JenkinsHook": "airflow.providers.jenkins.hooks.jenkins.JenkinsHook",
    },
    "jira_hook": {
        "JiraHook": "airflow.providers.atlassian.jira.hooks.jira.JiraHook",
    },
    "mongo_hook": {
        "MongoHook": "airflow.providers.mongo.hooks.mongo.MongoHook",
    },
    "openfaas_hook": {
        "OK_STATUS_CODE": "airflow.providers.openfaas.hooks.openfaas.OK_STATUS_CODE",
        "OpenFaasHook": "airflow.providers.openfaas.hooks.openfaas.OpenFaasHook",
        "requests": "airflow.providers.openfaas.hooks.openfaas.requests",
    },
    "opsgenie_alert_hook": {
        "OpsgenieAlertHook": "airflow.providers.opsgenie.hooks.opsgenie.OpsgenieAlertHook",
    },
    "pagerduty_hook": {
        "PagerdutyHook": "airflow.providers.pagerduty.hooks.pagerduty.PagerdutyHook",
    },
    "pinot_hook": {
        "PinotAdminHook": "airflow.providers.apache.pinot.hooks.pinot.PinotAdminHook",
        "PinotDbApiHook": "airflow.providers.apache.pinot.hooks.pinot.PinotDbApiHook",
    },
    "redis_hook": {
        "RedisHook": "airflow.providers.redis.hooks.redis.RedisHook",
    },
    "redshift_hook": {
        "RedshiftHook": "airflow.providers.amazon.aws.hooks.redshift_cluster.RedshiftHook",
    },
    "sagemaker_hook": {
        "LogState": "airflow.providers.amazon.aws.hooks.sagemaker.LogState",
        "Position": "airflow.providers.amazon.aws.hooks.sagemaker.Position",
        "SageMakerHook": "airflow.providers.amazon.aws.hooks.sagemaker.SageMakerHook",
        "argmin": "airflow.providers.amazon.aws.hooks.sagemaker.argmin",
        "secondary_training_status_changed": (
            "airflow.providers.amazon.aws.hooks.sagemaker.secondary_training_status_changed"
        ),
        "secondary_training_status_message": (
            "airflow.providers.amazon.aws.hooks.sagemaker.secondary_training_status_message"
        ),
    },
    "salesforce_hook": {
        "SalesforceHook": "airflow.providers.salesforce.hooks.salesforce.SalesforceHook",
        "pd": "airflow.providers.salesforce.hooks.salesforce.pd",
    },
    "segment_hook": {
        "SegmentHook": "airflow.providers.segment.hooks.segment.SegmentHook",
        "analytics": "airflow.providers.segment.hooks.segment.analytics",
    },
    "sftp_hook": {
        "SFTPHook": "airflow.providers.sftp.hooks.sftp.SFTPHook",
    },
    "slack_webhook_hook": {
        "SlackWebhookHook": "airflow.providers.slack.hooks.slack_webhook.SlackWebhookHook",
    },
    "snowflake_hook": {
        "SnowflakeHook": "airflow.providers.snowflake.hooks.snowflake.SnowflakeHook",
    },
    "spark_jdbc_hook": {
        "SparkJDBCHook": "airflow.providers.apache.spark.hooks.spark_jdbc.SparkJDBCHook",
    },
    "spark_sql_hook": {
        "SparkSqlHook": "airflow.providers.apache.spark.hooks.spark_sql.SparkSqlHook",
    },
    "spark_submit_hook": {
        "SparkSubmitHook": "airflow.providers.apache.spark.hooks.spark_submit.SparkSubmitHook",
    },
    "sqoop_hook": {
        "SqoopHook": "airflow.providers.apache.sqoop.hooks.sqoop.SqoopHook",
    },
    "ssh_hook": {
        "SSHHook": "airflow.providers.ssh.hooks.ssh.SSHHook",
    },
    "vertica_hook": {
        "VerticaHook": "airflow.providers.vertica.hooks.vertica.VerticaHook",
    },
    "wasb_hook": {
        "WasbHook": "airflow.providers.microsoft.azure.hooks.wasb.WasbHook",
    },
    "winrm_hook": {
        "WinRMHook": "airflow.providers.microsoft.winrm.hooks.winrm.WinRMHook",
    },
}

add_deprecated_classes(__deprecated_classes, __name__)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Triggerer command."""
from __future__ import annotations

from contextlib import contextmanager
from functools import partial
from multiprocessing import Process
from typing import Generator

from airflow import settings
from airflow.cli.commands.daemon_utils import run_command_with_daemon_option
from airflow.configuration import conf
from airflow.jobs.job import Job, run_job
from airflow.jobs.triggerer_job_runner import TriggererJobRunner
from airflow.utils import cli as cli_utils
from airflow.utils.providers_configuration_loader import providers_configuration_loaded
from airflow.utils.serve_logs import serve_logs


@contextmanager
def _serve_logs(skip_serve_logs: bool = False) -> Generator[None, None, None]:
    """Start serve_logs sub-process."""
    sub_proc = None
    if skip_serve_logs is False:
        port = conf.getint("logging", "trigger_log_server_port", fallback=8794)
        sub_proc = Process(target=partial(serve_logs, port=port))
        sub_proc.start()
    try:
        yield
    finally:
        if sub_proc:
            sub_proc.terminate()


def triggerer_run(skip_serve_logs: bool, capacity: int, triggerer_heartrate: float):
    with _serve_logs(skip_serve_logs):
        triggerer_job_runner = TriggererJobRunner(job=Job(heartrate=triggerer_heartrate), capacity=capacity)
        run_job(job=triggerer_job_runner.job, execute_callable=triggerer_job_runner._execute)


@cli_utils.action_cli
@providers_configuration_loaded
def triggerer(args):
    """Start Airflow Triggerer."""
    settings.MASK_SECRETS_IN_LOGS = True
    print(settings.HEADER)
    triggerer_heartrate = conf.getfloat("triggerer", "JOB_HEARTBEAT_SEC")

    run_command_with_daemon_option(
        args=args,
        process_name="triggerer",
        callback=lambda: triggerer_run(args.skip_serve_logs, args.capacity, triggerer_heartrate),
        should_setup_logging=True,
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Iterable, Optional

from pydantic import BaseModel as BaseModelPydantic, ConfigDict, PlainSerializer, PlainValidator
from typing_extensions import Annotated

from airflow.models import Operator
from airflow.models.baseoperator import BaseOperator
from airflow.models.taskinstance import TaskInstance
from airflow.serialization.pydantic.dag import DagModelPydantic
from airflow.serialization.pydantic.dag_run import DagRunPydantic
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.net import get_hostname
from airflow.utils.session import NEW_SESSION, provide_session
from airflow.utils.xcom import XCOM_RETURN_KEY

if TYPE_CHECKING:
    import pendulum
    from pydantic_core.core_schema import ValidationInfo
    from sqlalchemy.orm import Session

    from airflow.models.dagrun import DagRun
    from airflow.utils.context import Context
    from airflow.utils.state import DagRunState


def serialize_operator(x: Operator) -> dict:
    from airflow.serialization.serialized_objects import SerializedBaseOperator

    return SerializedBaseOperator.serialize_operator(x)


def validated_operator(x: dict[str, Any] | Operator, _info: ValidationInfo) -> Any:
    from airflow.models.baseoperator import BaseOperator
    from airflow.models.mappedoperator import MappedOperator
    from airflow.serialization.serialized_objects import SerializedBaseOperator

    if isinstance(x, BaseOperator) or isinstance(x, MappedOperator) or x is None:
        return x
    return SerializedBaseOperator.deserialize_operator(x)


PydanticOperator = Annotated[
    Operator,
    PlainValidator(validated_operator),
    PlainSerializer(serialize_operator, return_type=dict),
]


class TaskInstancePydantic(BaseModelPydantic, LoggingMixin):
    """Serializable representation of the TaskInstance ORM SqlAlchemyModel used by internal API."""

    task_id: str
    dag_id: str
    run_id: str
    map_index: int
    start_date: Optional[datetime]
    end_date: Optional[datetime]
    execution_date: Optional[datetime]
    duration: Optional[float]
    state: Optional[str]
    try_number: int
    _try_number: int
    max_tries: int
    hostname: str
    unixname: str
    job_id: Optional[int]
    pool: str
    pool_slots: int
    queue: str
    priority_weight: Optional[int]
    operator: str
    custom_operator_name: Optional[str]
    queued_dttm: Optional[str]
    queued_by_job_id: Optional[int]
    pid: Optional[int]
    executor_config: Any
    updated_at: Optional[datetime]
    external_executor_id: Optional[str]
    trigger_id: Optional[int]
    trigger_timeout: Optional[datetime]
    next_method: Optional[str]
    next_kwargs: Optional[dict]
    run_as_user: Optional[str]
    task: PydanticOperator
    test_mode: bool
    dag_run: Optional[DagRunPydantic]
    dag_model: Optional[DagModelPydantic]

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)

    def init_run_context(self, raw: bool = False) -> None:
        """Set the log context."""
        self.raw = raw
        self._set_context(self)

    def xcom_pull(
        self,
        task_ids: str | Iterable[str] | None = None,
        dag_id: str | None = None,
        key: str = XCOM_RETURN_KEY,
        include_prior_dates: bool = False,
        *,
        map_indexes: int | Iterable[int] | None = None,
        default: Any = None,
    ) -> Any:
        """
        Pull an XCom value for this task instance.

        TODO: make it works for AIP-44
        :param task_ids: task id or list of task ids, if None, the task_id of the current task is used
        :param dag_id: dag id, if None, the dag_id of the current task is used
        :param key: the key to identify the XCom value
        :param include_prior_dates: whether to include prior execution dates
        :param map_indexes: map index or list of map indexes, if None, the map_index of the current task
            is used
        :param default: the default value to return if the XCom value does not exist
        :return: Xcom value
        """
        return None

    @provide_session
    def xcom_push(
        self,
        key: str,
        value: Any,
        execution_date: datetime | None = None,
        session: Session = NEW_SESSION,
    ) -> None:
        """
        Push an XCom value for this task instance.

        TODO: make it works for AIP-44
        :param key: the key to identify the XCom value
        :param value: the value of the XCom
        :param execution_date: the execution date to push the XCom for
        """
        pass

    @provide_session
    def get_dagrun(self, session: Session = NEW_SESSION) -> DagRunPydantic:
        """
        Return the DagRun for this TaskInstance.

        :param session: SQLAlchemy ORM Session

        TODO: make it works for AIP-44

        :return: Pydantic serialized version of DaGrun
        """
        raise NotImplementedError()

    def _execute_task(self, context, task_orig):
        """
        Execute Task (optionally with a Timeout) and push Xcom results.

        :param context: Jinja2 context
        :param task_orig: origin task
        """
        from airflow.models.taskinstance import _execute_task

        return _execute_task(task_instance=self, context=context, task_orig=task_orig)

    @provide_session
    def refresh_from_db(self, session: Session = NEW_SESSION, lock_for_update: bool = False) -> None:
        """
        Refresh the task instance from the database based on the primary key.

        :param session: SQLAlchemy ORM Session
        :param lock_for_update: if True, indicates that the database should
            lock the TaskInstance (issuing a FOR UPDATE clause) until the
            session is committed.
        """
        from airflow.models.taskinstance import _refresh_from_db

        _refresh_from_db(task_instance=self, session=session, lock_for_update=lock_for_update)

    def set_duration(self) -> None:
        """Set task instance duration."""
        from airflow.models.taskinstance import _set_duration

        _set_duration(task_instance=self)

    @property
    def stats_tags(self) -> dict[str, str]:
        """Return task instance tags."""
        from airflow.models.taskinstance import _stats_tags

        return _stats_tags(task_instance=self)

    def clear_next_method_args(self) -> None:
        """Ensure we unset next_method and next_kwargs to ensure that any retries don't reuse them."""
        from airflow.models.taskinstance import _clear_next_method_args

        _clear_next_method_args(task_instance=self)

    def get_template_context(
        self,
        session: Session | None = None,
        ignore_param_exceptions: bool = True,
    ) -> Context:
        """
        Return TI Context.

        :param session: SQLAlchemy ORM Session
        :param ignore_param_exceptions: flag to suppress value exceptions while initializing the ParamsDict
        """
        from airflow.models.taskinstance import _get_template_context

        return _get_template_context(
            task_instance=self,
            session=session,
            ignore_param_exceptions=ignore_param_exceptions,
        )

    def is_eligible_to_retry(self):
        """Is task instance is eligible for retry."""
        from airflow.models.taskinstance import _is_eligible_to_retry

        return _is_eligible_to_retry(task_instance=self)

    @provide_session
    def handle_failure(
        self,
        error: None | str | Exception | KeyboardInterrupt,
        test_mode: bool | None = None,
        context: Context | None = None,
        force_fail: bool = False,
        session: Session = NEW_SESSION,
    ) -> None:
        """
        Handle Failure for a task instance.

        :param error: if specified, log the specific exception if thrown
        :param session: SQLAlchemy ORM Session
        :param test_mode: doesn't record success or failure in the DB if True
        :param context: Jinja2 context
        :param force_fail: if True, task does not retry
        """
        from airflow.models.taskinstance import _handle_failure

        _handle_failure(
            task_instance=self,
            error=error,
            session=session,
            test_mode=test_mode,
            context=context,
            force_fail=force_fail,
        )

    def refresh_from_task(self, task: Operator, pool_override: str | None = None) -> None:
        """
        Copy common attributes from the given task.

        :param task: The task object to copy from
        :param pool_override: Use the pool_override instead of task's pool
        """
        from airflow.models.taskinstance import _refresh_from_task

        _refresh_from_task(task_instance=self, task=task, pool_override=pool_override)

    @provide_session
    def get_previous_dagrun(
        self,
        state: DagRunState | None = None,
        session: Session | None = None,
    ) -> DagRun | None:
        """
        Return the DagRun that ran before this task instance's DagRun.

        :param state: If passed, it only take into account instances of a specific state.
        :param session: SQLAlchemy ORM Session.
        """
        from airflow.models.taskinstance import _get_previous_dagrun

        return _get_previous_dagrun(task_instance=self, state=state, session=session)

    @provide_session
    def get_previous_execution_date(
        self,
        state: DagRunState | None = None,
        session: Session = NEW_SESSION,
    ) -> pendulum.DateTime | None:
        """
        Return the execution date from property previous_ti_success.

        :param state: If passed, it only take into account instances of a specific state.
        :param session: SQLAlchemy ORM Session
        """
        from airflow.models.taskinstance import _get_previous_execution_date

        return _get_previous_execution_date(task_instance=self, state=state, session=session)

    def email_alert(self, exception, task: BaseOperator) -> None:
        """
        Send alert email with exception information.

        :param exception: the exception
        :param task: task related to the exception
        """
        from airflow.models.taskinstance import _email_alert

        _email_alert(task_instance=self, exception=exception, task=task)

    def get_email_subject_content(
        self, exception: BaseException, task: BaseOperator | None = None
    ) -> tuple[str, str, str]:
        """
        Get the email subject content for exceptions.

        :param exception: the exception sent in the email
        :param task:
        """
        from airflow.models.taskinstance import _get_email_subject_content

        return _get_email_subject_content(task_instance=self, exception=exception, task=task)

    @provide_session
    def get_previous_ti(
        self,
        state: DagRunState | None = None,
        session: Session = NEW_SESSION,
    ) -> TaskInstance | TaskInstancePydantic | None:
        """
        Return the task instance for the task that ran before this task instance.

        :param session: SQLAlchemy ORM Session
        :param state: If passed, it only take into account instances of a specific state.
        """
        from airflow.models.taskinstance import _get_previous_ti

        return _get_previous_ti(task_instance=self, state=state, session=session)

    @provide_session
    def check_and_change_state_before_execution(
        self,
        verbose: bool = True,
        ignore_all_deps: bool = False,
        ignore_depends_on_past: bool = False,
        wait_for_past_depends_before_skipping: bool = False,
        ignore_task_deps: bool = False,
        ignore_ti_state: bool = False,
        mark_success: bool = False,
        test_mode: bool = False,
        job_id: str | None = None,
        pool: str | None = None,
        external_executor_id: str | None = None,
        session: Session = NEW_SESSION,
    ) -> bool:
        return TaskInstance._check_and_change_state_before_execution(
            task_instance=self,
            verbose=verbose,
            ignore_all_deps=ignore_all_deps,
            ignore_depends_on_past=ignore_depends_on_past,
            wait_for_past_depends_before_skipping=wait_for_past_depends_before_skipping,
            ignore_task_deps=ignore_task_deps,
            ignore_ti_state=ignore_ti_state,
            mark_success=mark_success,
            test_mode=test_mode,
            hostname=get_hostname(),
            job_id=job_id,
            pool=pool,
            external_executor_id=external_executor_id,
            session=session,
        )

    @provide_session
    def schedule_downstream_tasks(self, session: Session = NEW_SESSION, max_tis_per_query: int | None = None):
        """
        Schedule downstream tasks of this task instance.

        :meta: private
        """
        return TaskInstance._schedule_downstream_tasks(
            ti=self, sessions=session, max_tis_per_query=max_tis_per_query
        )

    def command_as_list(
        self,
        mark_success: bool = False,
        ignore_all_deps: bool = False,
        ignore_task_deps: bool = False,
        ignore_depends_on_past: bool = False,
        wait_for_past_depends_before_skipping: bool = False,
        ignore_ti_state: bool = False,
        local: bool = False,
        pickle_id: int | None = None,
        raw: bool = False,
        job_id: str | None = None,
        pool: str | None = None,
        cfg_path: str | None = None,
    ) -> list[str]:
        """
        Return a command that can be executed anywhere where airflow is installed.

        This command is part of the message sent to executors by the orchestrator.
        """
        return TaskInstance._command_as_list(
            ti=self,
            mark_success=mark_success,
            ignore_all_deps=ignore_all_deps,
            ignore_task_deps=ignore_task_deps,
            ignore_depends_on_past=ignore_depends_on_past,
            wait_for_past_depends_before_skipping=wait_for_past_depends_before_skipping,
            ignore_ti_state=ignore_ti_state,
            local=local,
            pickle_id=pickle_id,
            raw=raw,
            job_id=job_id,
            pool=pool,
            cfg_path=cfg_path,
        )


TaskInstancePydantic.model_rebuild()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.utils.module_loading import qualname

serializers = ["deltalake.table.DeltaTable"]
deserializers = serializers
stringifiers = serializers

if TYPE_CHECKING:
    from airflow.serialization.serde import U

__version__ = 1


def serialize(o: object) -> tuple[U, str, int, bool]:
    from deltalake.table import DeltaTable

    if not isinstance(o, DeltaTable):
        return "", "", 0, False

    from airflow.models.crypto import get_fernet

    # we encrypt the information here until we have as part of the
    # storage options can have sensitive information
    fernet = get_fernet()
    properties: dict = {}
    for k, v in o._storage_options.items() if o._storage_options else {}:
        properties[k] = fernet.encrypt(v.encode("utf-8")).decode("utf-8")

    data = {
        "table_uri": o.table_uri,
        "version": o.version(),
        "storage_options": properties,
    }

    return data, qualname(o), __version__, True


def deserialize(classname: str, version: int, data: dict):
    from deltalake.table import DeltaTable

    from airflow.models.crypto import get_fernet

    if version > __version__:
        raise TypeError("serialized version is newer than class version")

    if classname == qualname(DeltaTable):
        fernet = get_fernet()
        properties = {}
        for k, v in data["storage_options"].items():
            properties[k] = fernet.decrypt(v.encode("utf-8")).decode("utf-8")

        if len(properties) == 0:
            storage_options = None
        else:
            storage_options = properties

        return DeltaTable(data["table_uri"], version=data["version"], storage_options=storage_options)

    raise TypeError(f"do not know how to deserialize {classname}")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from contextlib import suppress
from copy import copy
from logging import Logger
from typing import TYPE_CHECKING

from airflow.configuration import conf

if TYPE_CHECKING:
    from airflow.models.taskinstance import TaskInstance
    from airflow.utils.log.file_task_handler import FileTaskHandler

logger = logging.getLogger(__name__)


class TaskContextLogger:
    """
    Class for sending messages to task instance logs from outside task execution context.

    This is intended to be used mainly in exceptional circumstances, to give visibility into
    events related to task execution when otherwise there would be none.

    :meta private:
    """

    def __init__(self, component_name: str, call_site_logger: Logger | None = None):
        """
        Initialize the task context logger with the component name.

        :param component_name: the name of the component that will be used to identify the log messages
        :param call_site_logger: if provided, message will also be emitted through this logger
        """
        self.component_name = component_name
        self.task_handler = self._get_task_handler()
        self.enabled = self._should_enable()
        self.call_site_logger = call_site_logger

    def _should_enable(self) -> bool:
        if not conf.getboolean("logging", "enable_task_context_logger"):
            return False
        if not getattr(self.task_handler, "supports_task_context_logging", False):
            logger.warning("Task handler does not support task context logging")
            return False
        logger.info("Task context logging is enabled")
        return True

    @staticmethod
    def _get_task_handler() -> FileTaskHandler | None:
        """Returns the task handler that supports task context logging."""
        handlers = [
            handler
            for handler in logging.getLogger("airflow.task").handlers
            if getattr(handler, "supports_task_context_logging", False)
        ]
        if not handlers:
            return None
        h = handlers[0]
        if TYPE_CHECKING:
            assert isinstance(h, FileTaskHandler)
        return h

    def _log(self, level: int, msg: str, *args, ti: TaskInstance):
        """
        Emit a log message to the task instance logs.

        :param level: the log level
        :param msg: the message to relay to task context log
        :param ti: the task instance
        """
        if self.call_site_logger and self.call_site_logger.isEnabledFor(level=level):
            with suppress(Exception):
                self.call_site_logger.log(level, msg, *args)

        if not self.enabled:
            return

        if not self.task_handler:
            return

        task_handler = copy(self.task_handler)
        try:
            task_handler.set_context(ti, identifier=self.component_name)
            if hasattr(task_handler, "mark_end_on_close"):
                task_handler.mark_end_on_close = False
            filename, lineno, func, stackinfo = logger.findCaller()
            record = logging.LogRecord(
                self.component_name, level, filename, lineno, msg, args, None, func=func
            )
            task_handler.emit(record)
        finally:
            task_handler.close()

    def critical(self, msg: str, *args, ti: TaskInstance):
        """
        Emit a log message with level CRITICAL to the task instance logs.

        :param msg: the message to relay to task context log
        :param ti: the task instance
        """
        self._log(logging.CRITICAL, msg, *args, ti=ti)

    def fatal(self, msg: str, *args, ti: TaskInstance):
        """
        Emit a log message with level FATAL to the task instance logs.

        :param msg: the message to relay to task context log
        :param ti: the task instance
        """
        self._log(logging.FATAL, msg, *args, ti=ti)

    def error(self, msg: str, *args, ti: TaskInstance):
        """
        Emit a log message with level ERROR to the task instance logs.

        :param msg: the message to relay to task context log
        :param ti: the task instance
        """
        self._log(logging.ERROR, msg, *args, ti=ti)

    def warn(self, msg: str, *args, ti: TaskInstance):
        """
        Emit a log message with level WARN to the task instance logs.

        :param msg: the message to relay to task context log
        :param ti: the task instance
        """
        self._log(logging.WARN, msg, *args, ti=ti)

    def warning(self, msg: str, *args, ti: TaskInstance):
        """
        Emit a log message with level WARNING to the task instance logs.

        :param msg: the message to relay to task context log
        :param ti: the task instance
        """
        self._log(logging.WARNING, msg, *args, ti=ti)

    def info(self, msg: str, *args, ti: TaskInstance):
        """
        Emit a log message with level INFO to the task instance logs.

        :param msg: the message to relay to task context log
        :param ti: the task instance
        """
        self._log(logging.INFO, msg, *args, ti=ti)

    def debug(self, msg: str, *args, ti: TaskInstance):
        """
        Emit a log message with level DEBUG to the task instance logs.

        :param msg: the message to relay to task context log
        :param ti: the task instance
        """
        self._log(logging.DEBUG, msg, *args, ti=ti)

    def notset(self, msg: str, *args, ti: TaskInstance):
        """
        Emit a log message with level NOTSET to the task instance logs.

        :param msg: the message to relay to task context log
        :param ti: the task instance
        """
        self._log(logging.NOTSET, msg, *args, ti=ti)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Triggering DAG runs APIs."""
from __future__ import annotations

import json
from typing import TYPE_CHECKING

from airflow.exceptions import DagNotFound, DagRunAlreadyExists
from airflow.models import DagBag, DagModel, DagRun
from airflow.utils import timezone
from airflow.utils.state import DagRunState
from airflow.utils.types import DagRunType

if TYPE_CHECKING:
    from datetime import datetime


def _trigger_dag(
    dag_id: str,
    dag_bag: DagBag,
    run_id: str | None = None,
    conf: dict | str | None = None,
    execution_date: datetime | None = None,
    replace_microseconds: bool = True,
) -> list[DagRun | None]:
    """Triggers DAG run.

    :param dag_id: DAG ID
    :param dag_bag: DAG Bag model
    :param run_id: ID of the dag_run
    :param conf: configuration
    :param execution_date: date of execution
    :param replace_microseconds: whether microseconds should be zeroed
    :return: list of triggered dags
    """
    dag = dag_bag.get_dag(dag_id)  # prefetch dag if it is stored serialized

    if dag is None or dag_id not in dag_bag.dags:
        raise DagNotFound(f"Dag id {dag_id} not found")

    execution_date = execution_date or timezone.utcnow()

    if not timezone.is_localized(execution_date):
        raise ValueError("The execution_date should be localized")

    if replace_microseconds:
        execution_date = execution_date.replace(microsecond=0)

    if dag.default_args and "start_date" in dag.default_args:
        min_dag_start_date = dag.default_args["start_date"]
        if min_dag_start_date and execution_date < min_dag_start_date:
            raise ValueError(
                f"The execution_date [{execution_date.isoformat()}] should be >= start_date "
                f"[{min_dag_start_date.isoformat()}] from DAG's default_args"
            )
    logical_date = timezone.coerce_datetime(execution_date)

    data_interval = dag.timetable.infer_manual_data_interval(run_after=logical_date)
    run_id = run_id or dag.timetable.generate_run_id(
        run_type=DagRunType.MANUAL, logical_date=logical_date, data_interval=data_interval
    )
    dag_run = DagRun.find_duplicate(dag_id=dag_id, execution_date=execution_date, run_id=run_id)

    if dag_run:
        raise DagRunAlreadyExists(dag_run=dag_run, execution_date=execution_date, run_id=run_id)

    run_conf = None
    if conf:
        run_conf = conf if isinstance(conf, dict) else json.loads(conf)

    dag_runs = []
    dags_to_run = [dag, *dag.subdags]
    for _dag in dags_to_run:
        dag_run = _dag.create_dagrun(
            run_id=run_id,
            execution_date=execution_date,
            state=DagRunState.QUEUED,
            conf=run_conf,
            external_trigger=True,
            dag_hash=dag_bag.dags_hash.get(dag_id),
            data_interval=data_interval,
        )
        dag_runs.append(dag_run)

    return dag_runs


def trigger_dag(
    dag_id: str,
    run_id: str | None = None,
    conf: dict | str | None = None,
    execution_date: datetime | None = None,
    replace_microseconds: bool = True,
) -> DagRun | None:
    """Triggers execution of DAG specified by dag_id.

    :param dag_id: DAG ID
    :param run_id: ID of the dag_run
    :param conf: configuration
    :param execution_date: date of execution
    :param replace_microseconds: whether microseconds should be zeroed
    :return: first dag run triggered - even if more than one Dag Runs were triggered or None
    """
    dag_model = DagModel.get_current(dag_id)
    if dag_model is None:
        raise DagNotFound(f"Dag id {dag_id} not found in DagModel")

    dagbag = DagBag(dag_folder=dag_model.fileloc, read_dags_from_db=True)
    triggers = _trigger_dag(
        dag_id=dag_id,
        dag_bag=dagbag,
        run_id=run_id,
        conf=conf,
        execution_date=execution_date,
        replace_microseconds=replace_microseconds,
    )

    return triggers[0] if triggers else None



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from functools import partial
from typing import Any

from requests_kerberos import HTTPKerberosAuth

from airflow.api.auth.backend.kerberos_auth import (
    init_app as base_init_app,
    requires_authentication as base_requires_authentication,
)
from airflow.utils.airflow_flask_app import get_airflow_app

log = logging.getLogger(__name__)

CLIENT_AUTH: tuple[str, str] | Any | None = HTTPKerberosAuth(service="airflow")

init_app = base_init_app
requires_authentication = partial(
    base_requires_authentication, find_user=get_airflow_app().appbuilder.sm.find_user
)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Example DAG demonstrating the usage DAG params to model a trigger UI with a user form.

This example DAG generates greetings to a list of provided names in selected languages in the logs.
"""
from __future__ import annotations

import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from airflow.decorators import task
from airflow.models.dag import DAG
from airflow.models.param import Param
from airflow.utils.trigger_rule import TriggerRule

if TYPE_CHECKING:
    from airflow.models.dagrun import DagRun
    from airflow.models.taskinstance import TaskInstance

with DAG(
    dag_id=Path(__file__).stem,
    description=__doc__.partition(".")[0],
    doc_md=__doc__,
    schedule=None,
    start_date=datetime.datetime(2022, 3, 4),
    catchup=False,
    tags=["example_ui"],
    params={
        "names": Param(
            ["Linda", "Martha", "Thomas"],
            type="array",
            description="Define the list of names for which greetings should be generated in the logs."
            " Please have one name per line.",
            title="Names to greet",
        ),
        "english": Param(True, type="boolean", title="English"),
        "german": Param(True, type="boolean", title="German (Formal)"),
        "french": Param(True, type="boolean", title="French"),
    },
) as dag:

    @task(task_id="get_names")
    def get_names(**kwargs) -> list[str]:
        ti: TaskInstance = kwargs["ti"]
        dag_run: DagRun = ti.dag_run
        if "names" not in dag_run.conf:
            print("Uuups, no names given, was no UI used to trigger?")
            return []
        return dag_run.conf["names"]

    @task.branch(task_id="select_languages")
    def select_languages(**kwargs) -> list[str]:
        ti: TaskInstance = kwargs["ti"]
        dag_run: DagRun = ti.dag_run
        selected_languages = []
        for lang in ["english", "german", "french"]:
            if lang in dag_run.conf and dag_run.conf[lang]:
                selected_languages.append(f"generate_{lang}_greeting")
        return selected_languages

    @task(task_id="generate_english_greeting")
    def generate_english_greeting(name: str) -> str:
        return f"Hello {name}!"

    @task(task_id="generate_german_greeting")
    def generate_german_greeting(name: str) -> str:
        return f"Sehr geehrter Herr/Frau {name}."

    @task(task_id="generate_french_greeting")
    def generate_french_greeting(name: str) -> str:
        return f"Bonjour {name}!"

    @task(task_id="print_greetings", trigger_rule=TriggerRule.ALL_DONE)
    def print_greetings(greetings1, greetings2, greetings3) -> None:
        for g in greetings1 or []:
            print(g)
        for g in greetings2 or []:
            print(g)
        for g in greetings3 or []:
            print(g)
        if not (greetings1 or greetings2 or greetings3):
            print("sad, nobody to greet :-(")

    lang_select = select_languages()
    names = get_names()
    english_greetings = generate_english_greeting.expand(name=names)
    german_greetings = generate_german_greeting.expand(name=names)
    french_greetings = generate_french_greeting.expand(name=names)
    lang_select >> [english_greetings, german_greetings, french_greetings]
    results_print = print_greetings(english_greetings, german_greetings, french_greetings)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Make connection login/password TEXT

Revision ID: bd5dfbe21f88
Revises: f7bf2a57d0a6
Create Date: 2023-09-14 17:16:24.942390

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "bd5dfbe21f88"
down_revision = "f7bf2a57d0a6"
branch_labels = None
depends_on = None
airflow_version = "2.8.0"


def upgrade():
    """Apply Make connection login/password TEXT"""
    with op.batch_alter_table("connection", schema=None) as batch_op:
        batch_op.alter_column(
            "login", existing_type=sa.VARCHAR(length=500), type_=sa.Text(), existing_nullable=True
        )
        batch_op.alter_column(
            "password", existing_type=sa.VARCHAR(length=5000), type_=sa.Text(), existing_nullable=True
        )


def downgrade():
    """Unapply Make connection login/password TEXT"""
    with op.batch_alter_table("connection", schema=None) as batch_op:
        batch_op.alter_column(
            "password", existing_type=sa.Text(), type_=sa.VARCHAR(length=5000), existing_nullable=True
        )
        batch_op.alter_column(
            "login", existing_type=sa.Text(), type_=sa.VARCHAR(length=500), existing_nullable=True
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any

from airflow.template.templater import LiteralValue


def literal(value: Any) -> LiteralValue:
    """
    Wrap a value to ensure it is rendered as-is without applying Jinja templating to its contents.

    Designed for use in an operator's template field.

    :param value: The value to be rendered without templating
    """
    return LiteralValue(value)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum


@dataclass
class ConfigurationDetails:
    """Represents the details of a configuration."""

    section: str | None = None


@dataclass
class ConnectionDetails:
    """Represents the details of a connection."""

    conn_id: str | None = None


@dataclass
class DagDetails:
    """Represents the details of a DAG."""

    id: str | None = None


@dataclass
class DatasetDetails:
    """Represents the details of a dataset."""

    uri: str | None = None


@dataclass
class PoolDetails:
    """Represents the details of a pool."""

    name: str | None = None


@dataclass
class VariableDetails:
    """Represents the details of a variable."""

    key: str | None = None


class AccessView(Enum):
    """Enum of specific views the user tries to access."""

    CLUSTER_ACTIVITY = "CLUSTER_ACTIVITY"
    DOCS = "DOCS"
    JOBS = "JOBS"
    PLUGINS = "PLUGINS"
    PROVIDERS = "PROVIDERS"
    TRIGGERS = "TRIGGERS"
    WEBSITE = "WEBSITE"


class DagAccessEntity(Enum):
    """Enum of DAG entities the user tries to access."""

    AUDIT_LOG = "AUDIT_LOG"
    CODE = "CODE"
    DEPENDENCIES = "DEPENDENCIES"
    IMPORT_ERRORS = "IMPORT_ERRORS"
    RUN = "RUN"
    SLA_MISS = "SLA_MISS"
    TASK = "TASK"
    TASK_INSTANCE = "TASK_INSTANCE"
    TASK_RESCHEDULE = "TASK_RESCHEDULE"
    TASK_LOGS = "TASK_LOGS"
    WARNING = "WARNING"
    XCOM = "XCOM"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""add new field 'clear_number' to dagrun

Revision ID: 375a816bbbf4
Revises: 405de8318b3a
Create Date: 2023-09-05 19:27:30.531558

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "375a816bbbf4"
down_revision = "405de8318b3a"
branch_labels = None
depends_on = None
airflow_version = "2.8.0"


def upgrade():
    """Apply add cleared column to dagrun"""
    conn = op.get_bind()
    if conn.dialect.name == "mssql":
        with op.batch_alter_table("dag_run") as batch_op:
            batch_op.add_column(sa.Column("clear_number", sa.Integer, default=0))
            batch_op.alter_column("clear_number", existing_type=sa.Integer, nullable=False)
    else:
        with op.batch_alter_table("dag_run") as batch_op:
            batch_op.add_column(
                sa.Column(
                    "clear_number",
                    sa.Integer,
                    default=0,
                    nullable=False,
                    server_default="0",
                )
            )


def downgrade():
    """Unapply add cleared column to pool"""
    with op.batch_alter_table("dag_run") as batch_op:
        batch_op.drop_column("clear_number")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from airflow.auth.managers.base_auth_manager import ResourceMethod
    from airflow.auth.managers.models.base_user import BaseUser
    from airflow.auth.managers.models.resource_details import (
        ConnectionDetails,
        DagAccessEntity,
        DagDetails,
        PoolDetails,
        VariableDetails,
    )


class IsAuthorizedConnectionRequest(TypedDict, total=False):
    """Represent the parameters of ``is_authorized_connection`` API in the auth manager."""

    method: ResourceMethod
    details: ConnectionDetails | None
    user: BaseUser | None


class IsAuthorizedDagRequest(TypedDict, total=False):
    """Represent the parameters of ``is_authorized_dag`` API in the auth manager."""

    method: ResourceMethod
    access_entity: DagAccessEntity | None
    details: DagDetails | None
    user: BaseUser | None


class IsAuthorizedPoolRequest(TypedDict, total=False):
    """Represent the parameters of ``is_authorized_pool`` API in the auth manager."""

    method: ResourceMethod
    details: PoolDetails | None
    user: BaseUser | None


class IsAuthorizedVariableRequest(TypedDict, total=False):
    """Represent the parameters of ``is_authorized_variable`` API in the auth manager."""

    method: ResourceMethod
    details: VariableDetails | None
    user: BaseUser | None



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import functools
import gzip
import itertools
import json
import logging
from io import BytesIO
from typing import Callable, TypeVar, cast

import pendulum
from flask import after_this_request, request
from pendulum.parsing.exceptions import ParserError

from airflow.models import Log
from airflow.utils.log import secrets_masker
from airflow.utils.session import create_session
from airflow.www.extensions.init_auth_manager import get_auth_manager

T = TypeVar("T", bound=Callable)

logger = logging.getLogger(__name__)


def _mask_variable_fields(extra_fields):
    """
    Mask the 'val_content' field if 'key_content' is in the mask list.

    The variable requests values and args comes in this form:
    [('key', 'key_content'),('val', 'val_content'), ('description', 'description_content')]
    """
    result = []
    keyname = None
    for k, v in extra_fields:
        if k == "key":
            keyname = v
            result.append((k, v))
        elif keyname and k == "val":
            x = secrets_masker.redact(v, keyname)
            result.append((k, x))
            keyname = None
        else:
            result.append((k, v))
    return result


def _mask_connection_fields(extra_fields):
    """Mask connection fields."""
    result = []
    for k, v in extra_fields:
        if k == "extra":
            try:
                extra = json.loads(v)
                extra = [(k, secrets_masker.redact(v, k)) for k, v in extra.items()]
                result.append((k, json.dumps(dict(extra))))
            except json.JSONDecodeError:
                result.append((k, "Encountered non-JSON in `extra` field"))
        else:
            result.append((k, secrets_masker.redact(v, k)))
    return result


def action_logging(func: Callable | None = None, event: str | None = None) -> Callable[[T], T]:
    """Log user actions."""

    def log_action(f: T) -> T:
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            __tracebackhide__ = True  # Hide from pytest traceback.

            with create_session() as session:
                if not get_auth_manager().is_logged_in():
                    user = "anonymous"
                    user_display = ""
                else:
                    user = get_auth_manager().get_user_name()
                    user_display = get_auth_manager().get_user_display_name()

                fields_skip_logging = {"csrf_token", "_csrf_token"}
                extra_fields = [
                    (k, secrets_masker.redact(v, k))
                    for k, v in itertools.chain(request.values.items(multi=True), request.view_args.items())
                    if k not in fields_skip_logging
                ]
                if event and event.startswith("variable."):
                    extra_fields = _mask_variable_fields(extra_fields)
                if event and event.startswith("connection."):
                    extra_fields = _mask_connection_fields(extra_fields)

                params = {**request.values, **request.view_args}

                log = Log(
                    event=event or f.__name__,
                    task_instance=None,
                    owner=user,
                    owner_display_name=user_display,
                    extra=str(extra_fields),
                    task_id=params.get("task_id"),
                    dag_id=params.get("dag_id"),
                )

                if "execution_date" in request.values:
                    execution_date_value = request.values.get("execution_date")
                    try:
                        log.execution_date = pendulum.parse(execution_date_value, strict=False)
                    except ParserError:
                        logger.exception(
                            "Failed to parse execution_date from the request: %s", execution_date_value
                        )

                session.add(log)

            return f(*args, **kwargs)

        return cast(T, wrapper)

    if func:
        return log_action(func)
    return log_action


def gzipped(f: T) -> T:
    """Make a view compressed."""

    @functools.wraps(f)
    def view_func(*args, **kwargs):
        @after_this_request
        def zipper(response):
            accept_encoding = request.headers.get("Accept-Encoding", "")

            if "gzip" not in accept_encoding.lower():
                return response

            response.direct_passthrough = False

            if (
                response.status_code < 200
                or response.status_code >= 300
                or "Content-Encoding" in response.headers
            ):
                return response
            with BytesIO() as gzip_buffer:
                with gzip.GzipFile(mode="wb", fileobj=gzip_buffer) as gzip_file:
                    gzip_file.write(response.data)
                response.data = gzip_buffer.getvalue()
            response.headers["Content-Encoding"] = "gzip"
            response.headers["Vary"] = "Accept-Encoding"
            response.headers["Content-Length"] = len(response.data)

            return response

        return f(*args, **kwargs)

    return cast(T, view_func)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module provides helper code to make type annotation within Airflow codebase easier."""
from __future__ import annotations

__all__ = [
    "Literal",
    "ParamSpec",
    "Protocol",
    "TypedDict",
    "TypeGuard",
    "runtime_checkable",
]

import sys
from typing import Protocol, TypedDict, runtime_checkable

# Literal from typing module has various issues in different Python versions, see:
# - https://typing-extensions.readthedocs.io/en/latest/#Literal
# - bpo-45679: https://github.com/python/cpython/pull/29334
# - bpo-42345: https://github.com/python/cpython/pull/23294
# - bpo-42345: https://github.com/python/cpython/pull/23383
if sys.version_info >= (3, 10, 1) or (3, 9, 8) <= sys.version_info < (3, 10):
    from typing import Literal
else:
    from typing_extensions import Literal  # type: ignore[assignment]

if sys.version_info >= (3, 10):
    from typing import ParamSpec, TypeGuard
else:
    from typing_extensions import ParamSpec, TypeGuard



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Add owner_display_name to (Audit) Log table

Revision ID: f7bf2a57d0a6
Revises: 375a816bbbf4
Create Date: 2023-09-12 17:21:45.149658

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "f7bf2a57d0a6"
down_revision = "375a816bbbf4"
branch_labels = None
depends_on = None
airflow_version = "2.8.0"

TABLE_NAME = "log"


def upgrade():
    """Adds owner_display_name column to log"""
    with op.batch_alter_table(TABLE_NAME) as batch_op:
        batch_op.add_column(sa.Column("owner_display_name", sa.String(500)))


def downgrade():
    """Removes owner_display_name column from log"""
    with op.batch_alter_table(TABLE_NAME) as batch_op:
        batch_op.drop_column("owner_display_name")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import functools
import json
import logging
from typing import TYPE_CHECKING, Any, Callable

from flask import Response

from airflow.jobs.job import Job, most_recent_job
from airflow.serialization.serialized_objects import BaseSerialization
from airflow.utils.session import create_session

if TYPE_CHECKING:
    from airflow.api_connexion.types import APIResponse

log = logging.getLogger(__name__)


@functools.lru_cache
def _initialize_map() -> dict[str, Callable]:
    from airflow.dag_processing.manager import DagFileProcessorManager
    from airflow.dag_processing.processor import DagFileProcessor
    from airflow.models import Trigger, Variable, XCom
    from airflow.models.dag import DAG, DagModel
    from airflow.models.dagrun import DagRun
    from airflow.models.dagwarning import DagWarning
    from airflow.models.serialized_dag import SerializedDagModel
    from airflow.models.taskinstance import TaskInstance
    from airflow.secrets.metastore import MetastoreBackend

    functions: list[Callable] = [
        DagFileProcessor.update_import_errors,
        DagFileProcessor.manage_slas,
        DagFileProcessorManager.deactivate_stale_dags,
        DagModel.deactivate_deleted_dags,
        DagModel.get_paused_dag_ids,
        DagModel.get_current,
        DagFileProcessorManager.clear_nonexistent_import_errors,
        DagWarning.purge_inactive_dag_warnings,
        Job._add_to_db,
        Job._fetch_from_db,
        Job._kill,
        Job._update_heartbeat,
        Job._update_in_db,
        most_recent_job,
        MetastoreBackend._fetch_connection,
        MetastoreBackend._fetch_variable,
        XCom.get_value,
        XCom.get_one,
        XCom.get_many,
        XCom.clear,
        Variable.set,
        Variable.update,
        Variable.delete,
        DAG.fetch_callback,
        DAG.fetch_dagrun,
        DagRun.fetch_task_instances,
        DagRun.get_previous_dagrun,
        DagRun.get_previous_scheduled_dagrun,
        DagRun.fetch_task_instance,
        DagRun._get_log_template,
        SerializedDagModel.get_serialized_dag,
        TaskInstance._check_and_change_state_before_execution,
        TaskInstance.get_task_instance,
        TaskInstance.fetch_handle_failure_context,
        TaskInstance.save_to_db,
        TaskInstance._schedule_downstream_tasks,
        Trigger.from_object,
        Trigger.bulk_fetch,
        Trigger.clean_unused,
        Trigger.submit_event,
        Trigger.submit_failure,
        Trigger.ids_for_triggerer,
        Trigger.assign_unassigned,
    ]
    return {f"{func.__module__}.{func.__qualname__}": func for func in functions}


def internal_airflow_api(body: dict[str, Any]) -> APIResponse:
    """Handle Internal API /internal_api/v1/rpcapi endpoint."""
    log.debug("Got request")
    json_rpc = body.get("jsonrpc")
    if json_rpc != "2.0":
        log.error("Not jsonrpc-2.0 request.")
        return Response(response="Expected jsonrpc 2.0 request.", status=400)

    methods_map = _initialize_map()
    method_name = body.get("method")
    if method_name not in methods_map:
        log.error("Unrecognized method: %s.", method_name)
        return Response(response=f"Unrecognized method: {method_name}.", status=400)

    handler = methods_map[method_name]
    params = {}
    try:
        if body.get("params"):
            params_json = json.loads(str(body.get("params")))
            params = BaseSerialization.deserialize(params_json, use_pydantic_models=True)
    except Exception as e:
        log.error("Error when deserializing parameters for method: %s.", method_name)
        log.exception(e)
        return Response(response="Error deserializing parameters.", status=400)

    log.debug("Calling method %s.", method_name)
    try:
        # Session must be created there as it may be needed by serializer for lazy-loaded fields.
        with create_session() as session:
            output = handler(**params, session=session)
            output_json = BaseSerialization.serialize(output, use_pydantic_models=True)
            response = json.dumps(output_json) if output_json is not None else None
            return Response(response=response, headers={"Content-Type": "application/json"})
    except Exception as e:
        log.error("Error executing method: %s.", method_name)
        log.exception(e)
        return Response(response=f"Error executing method: {method_name}.", status=500)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module is deprecated.

Please use :mod:`airflow.auth.managers.fab.api.auth.backend.basic_auth` instead.
"""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Callable

import airflow.auth.managers.fab.api.auth.backend.basic_auth as fab_basic_auth
from airflow.exceptions import RemovedInAirflow3Warning

if TYPE_CHECKING:
    from airflow.auth.managers.fab.models import User

CLIENT_AUTH: tuple[str, str] | Any | None = None

warnings.warn(
    "This module is deprecated. Please use `airflow.auth.managers.fab.api.auth.backend.basic_auth` instead.",
    RemovedInAirflow3Warning,
    stacklevel=2,
)


def init_app(_):
    fab_basic_auth.init_app(_)


def auth_current_user() -> User | None:
    return fab_basic_auth.auth_current_user()


def requires_authentication(function: Callable):
    return fab_basic_auth.requires_authentication(function)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# fmt: off
"""Sensors."""
from __future__ import annotations

from airflow.utils.deprecation_tools import add_deprecated_classes

__deprecated_classes = {
    "base_sensor_operator": {
        "BaseSensorOperator": "airflow.sensors.base.BaseSensorOperator",
    },
    "date_time_sensor": {
        "DateTimeSensor": "airflow.sensors.date_time.DateTimeSensor",
    },
    "external_task_sensor": {
        "ExternalTaskMarker": "airflow.sensors.external_task.ExternalTaskMarker",
        "ExternalTaskSensor": "airflow.sensors.external_task.ExternalTaskSensor",
        "ExternalTaskSensorLink": "airflow.sensors.external_task.ExternalTaskSensorLink",
    },
    "hive_partition_sensor": {
        "HivePartitionSensor": "airflow.providers.apache.hive.sensors.hive_partition.HivePartitionSensor",
    },
    "http_sensor": {
        "HttpSensor": "airflow.providers.http.sensors.http.HttpSensor",
    },
    "metastore_partition_sensor": {
        "MetastorePartitionSensor": (
            "airflow.providers.apache.hive.sensors.metastore_partition.MetastorePartitionSensor"
        ),
    },
    "named_hive_partition_sensor": {
        "NamedHivePartitionSensor": (
            "airflow.providers.apache.hive.sensors.named_hive_partition.NamedHivePartitionSensor"
        ),
    },
    "s3_key_sensor": {
        "S3KeySensor": "airflow.providers.amazon.aws.sensors.s3.S3KeySensor",
    },
    "sql": {
        "SqlSensor": "airflow.providers.common.sql.sensors.sql.SqlSensor",
    },
    "sql_sensor": {
        "SqlSensor": "airflow.providers.common.sql.sensors.sql.SqlSensor",
    },
    "time_delta_sensor": {
        "TimeDeltaSensor": "airflow.sensors.time_delta.TimeDeltaSensor",
    },
    "web_hdfs_sensor": {
        "WebHdfsSensor": "airflow.providers.apache.hdfs.sensors.web_hdfs.WebHdfsSensor",
    },
}

add_deprecated_classes(__deprecated_classes, __name__)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.security.permissions import (
    ACTION_CAN_ACCESS_MENU,
    ACTION_CAN_CREATE,
    ACTION_CAN_DELETE,
    ACTION_CAN_EDIT,
    ACTION_CAN_READ,
)

if TYPE_CHECKING:
    from airflow.auth.managers.base_auth_manager import ResourceMethod

# Convert methods to FAB action name
_MAP_METHOD_NAME_TO_FAB_ACTION_NAME: dict[ResourceMethod, str] = {
    "POST": ACTION_CAN_CREATE,
    "GET": ACTION_CAN_READ,
    "PUT": ACTION_CAN_EDIT,
    "DELETE": ACTION_CAN_DELETE,
}


def get_fab_action_from_method_map():
    """Returns the map associating a method to a FAB action."""
    return _MAP_METHOD_NAME_TO_FAB_ACTION_NAME


def get_method_from_fab_action_map():
    """Returns the map associating a FAB action to a method."""
    return {
        **{v: k for k, v in _MAP_METHOD_NAME_TO_FAB_ACTION_NAME.items()},
        ACTION_CAN_ACCESS_MENU: "GET",
    }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from abc import abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Container, Literal, Sequence

from sqlalchemy import select

from airflow.auth.managers.models.resource_details import (
    DagDetails,
)
from airflow.exceptions import AirflowException
from airflow.models import DagModel
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.session import NEW_SESSION, provide_session

if TYPE_CHECKING:
    from connexion import FlaskApi
    from sqlalchemy.orm import Session

    from airflow.auth.managers.models.base_user import BaseUser
    from airflow.auth.managers.models.batch_apis import (
        IsAuthorizedConnectionRequest,
        IsAuthorizedDagRequest,
        IsAuthorizedPoolRequest,
        IsAuthorizedVariableRequest,
    )
    from airflow.auth.managers.models.resource_details import (
        AccessView,
        ConfigurationDetails,
        ConnectionDetails,
        DagAccessEntity,
        DatasetDetails,
        PoolDetails,
        VariableDetails,
    )
    from airflow.cli.cli_config import CLICommand
    from airflow.www.extensions.init_appbuilder import AirflowAppBuilder
    from airflow.www.security_manager import AirflowSecurityManagerV2

ResourceMethod = Literal["GET", "POST", "PUT", "DELETE"]


class BaseAuthManager(LoggingMixin):
    """
    Class to derive in order to implement concrete auth managers.

    Auth managers are responsible for any user management related operation such as login, logout, authz, ...

    :param appbuilder: the flask app builder
    """

    def __init__(self, appbuilder: AirflowAppBuilder) -> None:
        super().__init__()
        self.appbuilder = appbuilder

    @staticmethod
    def get_cli_commands() -> list[CLICommand]:
        """Vends CLI commands to be included in Airflow CLI.

        Override this method to expose commands via Airflow CLI to manage this auth manager.
        """
        return []

    def get_api_endpoints(self) -> None | FlaskApi:
        """Return API endpoint(s) definition for the auth manager."""
        return None

    def get_user_name(self) -> str:
        """Return the username associated to the user in session."""
        user = self.get_user()
        if not user:
            self.log.error("Calling 'get_user_name()' but the user is not signed in.")
            raise AirflowException("The user must be signed in.")
        return user.get_name()

    def get_user_display_name(self) -> str:
        """Return the user's display name associated to the user in session."""
        return self.get_user_name()

    @abstractmethod
    def get_user(self) -> BaseUser | None:
        """Return the user associated to the user in session."""

    def get_user_id(self) -> str:
        """Return the user ID associated to the user in session."""
        user = self.get_user()
        if not user:
            self.log.error("Calling 'get_user_id()' but the user is not signed in.")
            raise AirflowException("The user must be signed in.")
        return str(user.get_id())

    def init(self) -> None:
        """
        Run operations when Airflow is initializing.

        By default, do nothing.
        """

    @abstractmethod
    def is_logged_in(self) -> bool:
        """Return whether the user is logged in."""

    @abstractmethod
    def is_authorized_configuration(
        self,
        *,
        method: ResourceMethod,
        details: ConfigurationDetails | None = None,
        user: BaseUser | None = None,
    ) -> bool:
        """
        Return whether the user is authorized to perform a given action on configuration.

        :param method: the method to perform
        :param details: optional details about the configuration
        :param user: the user to perform the action on. If not provided (or None), it uses the current user
        """

    @abstractmethod
    def is_authorized_cluster_activity(
        self,
        *,
        method: ResourceMethod,
        user: BaseUser | None = None,
    ) -> bool:
        """
        Return whether the user is authorized to perform a given action on the cluster activity.

        :param method: the method to perform
        :param user: the user to perform the action on. If not provided (or None), it uses the current user
        """

    @abstractmethod
    def is_authorized_connection(
        self,
        *,
        method: ResourceMethod,
        details: ConnectionDetails | None = None,
        user: BaseUser | None = None,
    ) -> bool:
        """
        Return whether the user is authorized to perform a given action on a connection.

        :param method: the method to perform
        :param details: optional details about the connection
        :param user: the user to perform the action on. If not provided (or None), it uses the current user
        """

    @abstractmethod
    def is_authorized_dag(
        self,
        *,
        method: ResourceMethod,
        access_entity: DagAccessEntity | None = None,
        details: DagDetails | None = None,
        user: BaseUser | None = None,
    ) -> bool:
        """
        Return whether the user is authorized to perform a given action on a DAG.

        :param method: the method to perform
        :param access_entity: the kind of DAG information the authorization request is about.
            If not provided, the authorization request is about the DAG itself
        :param details: optional details about the DAG
        :param user: the user to perform the action on. If not provided (or None), it uses the current user
        """

    @abstractmethod
    def is_authorized_dataset(
        self,
        *,
        method: ResourceMethod,
        details: DatasetDetails | None = None,
        user: BaseUser | None = None,
    ) -> bool:
        """
        Return whether the user is authorized to perform a given action on a dataset.

        :param method: the method to perform
        :param details: optional details about the dataset
        :param user: the user to perform the action on. If not provided (or None), it uses the current user
        """

    @abstractmethod
    def is_authorized_pool(
        self,
        *,
        method: ResourceMethod,
        details: PoolDetails | None = None,
        user: BaseUser | None = None,
    ) -> bool:
        """
        Return whether the user is authorized to perform a given action on a pool.

        :param method: the method to perform
        :param details: optional details about the pool
        :param user: the user to perform the action on. If not provided (or None), it uses the current user
        """

    @abstractmethod
    def is_authorized_variable(
        self,
        *,
        method: ResourceMethod,
        details: VariableDetails | None = None,
        user: BaseUser | None = None,
    ) -> bool:
        """
        Return whether the user is authorized to perform a given action on a variable.

        :param method: the method to perform
        :param details: optional details about the variable
        :param user: the user to perform the action on. If not provided (or None), it uses the current user
        """

    @abstractmethod
    def is_authorized_view(
        self,
        *,
        access_view: AccessView,
        user: BaseUser | None = None,
    ) -> bool:
        """
        Return whether the user is authorized to access a read-only state of the installation.

        :param access_view: the specific read-only view/state the authorization request is about.
        :param user: the user to perform the action on. If not provided (or None), it uses the current user
        """

    def is_authorized_custom_view(
        self, *, fab_action_name: str, fab_resource_name: str, user: BaseUser | None = None
    ):
        """
        Return whether the user is authorized to perform a given action on a custom view.

        A custom view is a view defined as part of the auth manager. This view is then only available when
        the auth manager is used as part of the environment.

        By default, it throws an exception because auth managers do not define custom views by default.
        If an auth manager defines some custom views, it needs to override this method.

        :param fab_action_name: the name of the FAB action defined in the view in ``base_permissions``
        :param fab_resource_name: the name of the FAB resource defined in the view in
            ``class_permission_name``
        :param user: the user to perform the action on. If not provided (or None), it uses the current user
        """
        raise AirflowException(f"The resource `{fab_resource_name}` does not exist in the environment.")

    def batch_is_authorized_dag(
        self,
        requests: Sequence[IsAuthorizedDagRequest],
    ) -> bool:
        """
        Batch version of ``is_authorized_dag``.

        By default, calls individually the ``is_authorized_dag`` API on each item in the list of requests.
        Can lead to some poor performance. It is recommended to override this method in the auth manager
        implementation to provide a more efficient implementation.

        :param requests: a list of requests containing the parameters for ``is_authorized_dag``
        """
        return all(
            self.is_authorized_dag(
                method=request["method"],
                access_entity=request.get("access_entity"),
                details=request.get("details"),
                user=request.get("user"),
            )
            for request in requests
        )

    def batch_is_authorized_connection(
        self,
        requests: Sequence[IsAuthorizedConnectionRequest],
    ) -> bool:
        """
        Batch version of ``is_authorized_connection``.

        By default, calls individually the ``is_authorized_connection`` API on each item in the list of
        requests. Can lead to some poor performance. It is recommended to override this method in the auth
        manager implementation to provide a more efficient implementation.

        :param requests: a list of requests containing the parameters for ``is_authorized_connection``
        """
        return all(
            self.is_authorized_connection(
                method=request["method"], details=request.get("details"), user=request.get("user")
            )
            for request in requests
        )

    def batch_is_authorized_pool(
        self,
        requests: Sequence[IsAuthorizedPoolRequest],
    ) -> bool:
        """
        Batch version of ``is_authorized_pool``.

        By default, calls individually the ``is_authorized_pool`` API on each item in the list of
        requests. Can lead to some poor performance. It is recommended to override this method in the auth
        manager implementation to provide a more efficient implementation.

        :param requests: a list of requests containing the parameters for ``is_authorized_pool``
        """
        return all(
            self.is_authorized_pool(
                method=request["method"], details=request.get("details"), user=request.get("user")
            )
            for request in requests
        )

    def batch_is_authorized_variable(
        self,
        requests: Sequence[IsAuthorizedVariableRequest],
    ) -> bool:
        """
        Batch version of ``is_authorized_variable``.

        By default, calls individually the ``is_authorized_variable`` API on each item in the list of
        requests. Can lead to some poor performance. It is recommended to override this method in the auth
        manager implementation to provide a more efficient implementation.

        :param requests: a list of requests containing the parameters for ``is_authorized_variable``
        """
        return all(
            self.is_authorized_variable(
                method=request["method"], details=request.get("details"), user=request.get("user")
            )
            for request in requests
        )

    @provide_session
    def get_permitted_dag_ids(
        self,
        *,
        methods: Container[ResourceMethod] | None = None,
        user=None,
        session: Session = NEW_SESSION,
    ) -> set[str]:
        """
        Get readable or writable DAGs for user.

        By default, reads all the DAGs and check individually if the user has permissions to access the DAG.
        Can lead to some poor performance. It is recommended to override this method in the auth manager
        implementation to provide a more efficient implementation.
        """
        if not methods:
            methods = ["PUT", "GET"]

        dag_ids = {dag.dag_id for dag in session.execute(select(DagModel.dag_id))}

        if ("GET" in methods and self.is_authorized_dag(method="GET", user=user)) or (
            "PUT" in methods and self.is_authorized_dag(method="PUT", user=user)
        ):
            # If user is authorized to read/edit all DAGs, return all DAGs
            return dag_ids

        def _is_permitted_dag_id(method: ResourceMethod, methods: Container[ResourceMethod], dag_id: str):
            return method in methods and self.is_authorized_dag(
                method=method, details=DagDetails(id=dag_id), user=user
            )

        return {
            dag_id
            for dag_id in dag_ids
            if _is_permitted_dag_id("GET", methods, dag_id) or _is_permitted_dag_id("PUT", methods, dag_id)
        }

    @abstractmethod
    def get_url_login(self, **kwargs) -> str:
        """Return the login page url."""

    @abstractmethod
    def get_url_logout(self) -> str:
        """Return the logout page url."""

    def get_url_user_profile(self) -> str | None:
        """
        Return the url to a page displaying info about the current user.

        By default, return None.
        """
        return None

    @cached_property
    def security_manager(self) -> AirflowSecurityManagerV2:
        """
        Return the security manager.

        By default, Airflow comes with the default security manager
        ``airflow.www.security_manager.AirflowSecurityManagerV2``. The auth manager might need to extend this
        default security manager for its own purposes.

        By default, return the default AirflowSecurityManagerV2.
        """
        from airflow.www.security_manager import AirflowSecurityManagerV2

        return AirflowSecurityManagerV2(self.appbuilder)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Add processor_subdir to ImportError.

Revision ID: 10b52ebd31f7
Revises: bd5dfbe21f88
Create Date: 2023-11-29 16:54:48.101834

"""

import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "10b52ebd31f7"
down_revision = "bd5dfbe21f88"
branch_labels = None
depends_on = None
airflow_version = "2.8.0"


def upgrade():
    """Apply Add processor_subdir to ImportError."""
    conn = op.get_bind()

    with op.batch_alter_table("import_error") as batch_op:
        if conn.dialect.name == "mysql":
            batch_op.add_column(sa.Column("processor_subdir", sa.Text(length=2000), nullable=True))
        else:
            batch_op.add_column(sa.Column("processor_subdir", sa.String(length=2000), nullable=True))


def downgrade():
    """Unapply Add processor_subdir to ImportError."""
    conn = op.get_bind()
    with op.batch_alter_table("import_error", schema=None) as batch_op:
        batch_op.drop_column("processor_subdir")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
import functools
import os
import shutil
import typing
from pathlib import PurePath
from urllib.parse import urlsplit

from fsspec.core import split_protocol
from fsspec.utils import stringify_path
from upath.implementations.cloud import CloudPath, _CloudAccessor
from upath.registry import get_upath_class

from airflow.io.store import attach
from airflow.io.utils.stat import stat_result

if typing.TYPE_CHECKING:
    from urllib.parse import SplitResult

    from fsspec import AbstractFileSystem


PT = typing.TypeVar("PT", bound="ObjectStoragePath")

default = "file"


class _AirflowCloudAccessor(_CloudAccessor):
    __slots__ = ("_store",)

    def __init__(
        self,
        parsed_url: SplitResult | None,
        conn_id: str | None = None,
        **kwargs: typing.Any,
    ) -> None:
        if parsed_url and parsed_url.scheme:
            self._store = attach(parsed_url.scheme, conn_id)
        else:
            self._store = attach("file", conn_id)

    @property
    def _fs(self) -> AbstractFileSystem:
        return self._store.fs

    def __eq__(self, other):
        return isinstance(other, _AirflowCloudAccessor) and self._store == other._store


class ObjectStoragePath(CloudPath):
    """A path-like object for object storage."""

    _accessor: _AirflowCloudAccessor

    __version__: typing.ClassVar[int] = 1

    _default_accessor = _AirflowCloudAccessor

    sep: typing.ClassVar[str] = "/"
    root_marker: typing.ClassVar[str] = "/"

    _bucket: str
    _key: str
    _protocol: str
    _hash: int | None

    __slots__ = (
        "_bucket",
        "_key",
        "_conn_id",
        "_protocol",
        "_hash",
    )

    def __new__(
        cls: type[PT],
        *args: str | os.PathLike,
        scheme: str | None = None,
        conn_id: str | None = None,
        **kwargs: typing.Any,
    ) -> PT:
        args_list = list(args)

        if args_list:
            other = args_list.pop(0) or "."
        else:
            other = "."

        if isinstance(other, PurePath):
            _cls: typing.Any = type(other)
            drv, root, parts = _cls._parse_args(args_list)
            drv, root, parts = _cls._flavour.join_parsed_parts(
                other._drv,  # type: ignore[attr-defined]
                other._root,  # type: ignore[attr-defined]
                other._parts,  # type: ignore[attr-defined]
                drv,
                root,
                parts,  # type: ignore
            )

            _kwargs = getattr(other, "_kwargs", {})
            _url = getattr(other, "_url", None)
            other_kwargs = _kwargs.copy()
            if _url and _url.scheme:
                other_kwargs["url"] = _url
            new_kwargs = _kwargs.copy()
            new_kwargs.update(kwargs)

            return _cls(_cls._format_parsed_parts(drv, root, parts, **other_kwargs), **new_kwargs)

        url = stringify_path(other)
        parsed_url: SplitResult = urlsplit(url)

        if scheme:  # allow override of protocol
            parsed_url = parsed_url._replace(scheme=scheme)

        if not parsed_url.path:  # ensure path has root
            parsed_url = parsed_url._replace(path="/")

        if not parsed_url.scheme and not split_protocol(url)[0]:
            args_list.insert(0, url)
        else:
            args_list.insert(0, parsed_url.path)

        # This matches the parsing logic in urllib.parse; see:
        # https://github.com/python/cpython/blob/46adf6b701c440e047abf925df9a75a/Lib/urllib/parse.py#L194-L203
        userinfo, have_info, hostinfo = parsed_url.netloc.rpartition("@")
        if have_info:
            conn_id = conn_id or userinfo or None
            parsed_url = parsed_url._replace(netloc=hostinfo)

        return cls._from_parts(args_list, url=parsed_url, conn_id=conn_id, **kwargs)  # type: ignore

    @functools.lru_cache
    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: typing.Any) -> bool:
        return self.samestore(other) and str(self) == str(other)

    def samestore(self, other: typing.Any) -> bool:
        return isinstance(other, ObjectStoragePath) and self._accessor == other._accessor

    @property
    def container(self) -> str:
        return self.bucket

    @property
    def bucket(self) -> str:
        if self._url:
            return self._url.netloc
        else:
            return ""

    @property
    def key(self) -> str:
        if self._url:
            return self._url.path
        else:
            return ""

    def stat(self) -> stat_result:  # type: ignore[override]
        """Call ``stat`` and return the result."""
        return stat_result(
            self._accessor.stat(self),
            protocol=self.protocol,
            conn_id=self._accessor._store.conn_id,
        )

    def samefile(self, other_path: typing.Any) -> bool:
        """Return whether other_path is the same or not as this file."""
        if not isinstance(other_path, ObjectStoragePath):
            return False

        st = self.stat()
        other_st = other_path.stat()

        return (
            st["protocol"] == other_st["protocol"]
            and st["conn_id"] == other_st["conn_id"]
            and st["ino"] == other_st["ino"]
        )

    def _scandir(self):
        # Emulate os.scandir(), which returns an object that can be used as a
        # context manager.
        return contextlib.nullcontext(self.iterdir())

    def replace(self, target) -> ObjectStoragePath:
        """
        Rename this path to the target path, overwriting if that path exists.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        return self.rename(target, overwrite=True)

    @classmethod
    def cwd(cls):
        if cls is ObjectStoragePath:
            return get_upath_class("").cwd()
        else:
            raise NotImplementedError

    @classmethod
    def home(cls):
        if cls is ObjectStoragePath:
            return get_upath_class("").home()
        else:
            raise NotImplementedError

    # EXTENDED OPERATIONS

    def ukey(self) -> str:
        """Hash of file properties, to tell if it has changed."""
        return self.fs.ukey(self.path)

    def checksum(self) -> int:
        """Return the checksum of the file at this path."""
        # we directly access the fs here to avoid changing the abstract interface
        return self.fs.checksum(self.path)

    def read_block(self, offset: int, length: int, delimiter=None):
        r"""Read a block of bytes.

        Starting at ``offset`` of the file, read ``length`` bytes. If
        ``delimiter`` is set then we ensure that the read starts and stops at
        delimiter boundaries that follow the locations ``offset`` and ``offset
        + length``. If ``offset`` is zero then we start at zero. The
        bytestring returned WILL include the end delimiter string.

        If offset+length is beyond the eof, reads to eof.

        :param offset: int
                      Byte offset to start read
        :param length: int
                      Number of bytes to read. If None, read to the end.
        :param delimiter: bytes (optional)
                        Ensure reading starts and stops at delimiter bytestring

        Examples
        --------
        >>> read_block(0, 13)
        b'Alice, 100\\nBo'
        >>> read_block(0, 13, delimiter=b'\\n')
        b'Alice, 100\\nBob, 200\\n'

        Use ``length=None`` to read to the end of the file.
        >>> read_block(0, None, delimiter=b'\\n')
        b'Alice, 100\\nBob, 200\\nCharlie, 300'

        See Also
        --------
        :func:`fsspec.utils.read_block`
        """
        return self.fs.read_block(self.path, offset=offset, length=length, delimiter=delimiter)

    def sign(self, expiration: int = 100, **kwargs):
        """Create a signed URL representing the given path.

        Some implementations allow temporary URLs to be generated, as a
        way of delegating credentials.

        :param path: str
                     The path on the filesystem
        :param expiration: int
                          Number of seconds to enable the URL for (if supported)

        :returns URL: str
                     The signed URL

        :raises NotImplementedError: if the method is not implemented for a store
        """
        return self.fs.sign(self.path, expiration=expiration, **kwargs)

    def size(self) -> int:
        """Size in bytes of the file at this path."""
        return self.fs.size(self.path)

    def _cp_file(self, dst: ObjectStoragePath, **kwargs):
        """Copy a single file from this path to another location by streaming the data."""
        # create the directory or bucket if required
        if dst.key.endswith(self.sep) or not dst.key:
            dst.mkdir(exist_ok=True, parents=True)
            dst = dst / self.key
        elif dst.is_dir():
            dst = dst / self.key

        # streaming copy
        with self.open("rb") as f1, dst.open("wb") as f2:
            # make use of system dependent buffer size
            shutil.copyfileobj(f1, f2, **kwargs)

    def copy(self, dst: str | ObjectStoragePath, recursive: bool = False, **kwargs) -> None:
        """Copy file(s) from this path to another location.

        For remote to remote copies, the key used for the destination will be the same as the source.
        So that s3://src_bucket/foo/bar will be copied to gcs://dst_bucket/foo/bar and not
        gcs://dst_bucket/bar.

        :param dst: Destination path
        :param recursive: If True, copy directories recursively.

        kwargs: Additional keyword arguments to be passed to the underlying implementation.
        """
        if isinstance(dst, str):
            dst = ObjectStoragePath(dst)

        # same -> same
        if self.samestore(dst):
            self.fs.copy(self.path, dst.path, recursive=recursive, **kwargs)
            return

        # use optimized path for local -> remote or remote -> local
        if self.protocol == "file":
            dst.fs.put(self.path, dst.path, recursive=recursive, **kwargs)
            return

        if dst.protocol == "file":
            self.fs.get(self.path, dst.path, recursive=recursive, **kwargs)
            return

        if not self.exists():
            raise FileNotFoundError(f"{self} does not exist")

        # remote dir -> remote dir
        if self.is_dir():
            if dst.is_file():
                raise ValueError("Cannot copy directory to a file.")

            dst.mkdir(exist_ok=True, parents=True)

            out = self.fs.expand_path(self.path, recursive=True, **kwargs)

            for path in out:
                # this check prevents one extra call to is_dir() as
                # glob returns self as well
                if path == self.path:
                    continue

                src_obj = ObjectStoragePath(path, conn_id=self._accessor._store.conn_id)

                # skip directories, empty directories will not be created
                if src_obj.is_dir():
                    continue

                src_obj._cp_file(dst)

            return

        # remote file -> remote dir
        self._cp_file(dst, **kwargs)

    def move(self, path: str | ObjectStoragePath, recursive: bool = False, **kwargs) -> None:
        """Move file(s) from this path to another location.

        :param path: Destination path
        :param recursive: bool
                         If True, move directories recursively.

        kwargs: Additional keyword arguments to be passed to the underlying implementation.
        """
        if isinstance(path, str):
            path = ObjectStoragePath(path)

        if self.samestore(path):
            return self.fs.move(self.path, path.path, recursive=recursive, **kwargs)

        # non-local copy
        self.copy(path, recursive=recursive, **kwargs)
        self.unlink()

    def serialize(self) -> dict[str, typing.Any]:
        _kwargs = self._kwargs.copy()
        conn_id = _kwargs.pop("conn_id", None)

        return {
            "path": str(self),
            "conn_id": conn_id,
            "kwargs": _kwargs,
        }

    @classmethod
    def deserialize(cls, data: dict, version: int) -> ObjectStoragePath:
        if version > cls.__version__:
            raise ValueError(f"Cannot deserialize version {version} with version {cls.__version__}.")

        _kwargs = data.pop("kwargs")
        path = data.pop("path")
        conn_id = data.pop("conn_id", None)

        return ObjectStoragePath(path, conn_id=conn_id, **_kwargs)
